import json
from textwrap import indent
from typing import List

from pyspark.sql.utils import AnalysisException

from spetlr.configurator import Configurator
from spetlr.delta import DeltaHandle
from spetlr.delta.delta_handle import DeltaHandleInvalidName
from spetlr.sql import SqlExecutor
from spetlr.testutils import CleanupTestDatabases


def get_table_ids_to_check():
    print("Remember to initialize the configurator first!")
    c = Configurator()
    table_ids_to_check = []
    for key in c._raw_resource_details.keys():
        if c.table_property(key, "delete_on_delta_schema_mismatch", False):
            table_ids_to_check.append(key)
    return table_ids_to_check


def DeleteMismatchedSchemas(
    table_ids_to_check: List[str] = None,
    spark_executor: SqlExecutor = None,
    sql_files_pattern: str = "*",
):
    """For the following tables, if the production schema does not match
    the configured schema, delete the table.

    It is the responsibility of the developer to only add tables here where the
    code has the property that it can rebuild dropped tables.

    Tables can be flagged by using ´delete_on_delta_schema_mismatch´
    as Configurator property.

    """

    if table_ids_to_check is None:
        table_ids_to_check = get_table_ids_to_check()

    executor = spark_executor

    # first gather the configured schemas from the test databases
    print("Remember to initialize the configurator first!")
    configurator = Configurator()
    configurator.set_debug()
    executor.execute_sql_file(sql_files_pattern)

    schemas = {}
    for tbl_id in table_ids_to_check:
        try:
            schemas[tbl_id] = DeltaHandle.from_tc(tbl_id).read().schema
        except DeltaHandleInvalidName:
            print(f"Not a valid delta handle {tbl_id}")
            continue

    # cleanup
    CleanupTestDatabases()

    # now get the production schemas from the production tables
    affected_keys = []
    configurator.set_prod()
    for tbl_id in table_ids_to_check:
        try:
            prod_schema = DeltaHandle.from_tc(tbl_id).read().schema
        except AnalysisException:
            print(f"Exception in reading production version of table id {tbl_id}")
            continue
        except DeltaHandleInvalidName:
            print(f"Not a valid delta handle {tbl_id}")
            continue
        if prod_schema != schemas[tbl_id]:
            print(f"Schema mismatch detected for table id {tbl_id}.")
            print(
                "  Production table schema:",
                indent(json.dumps(prod_schema.jsonValue(), indent=4), "  "),
            )
            print(
                "  Configured table schema:",
                indent(json.dumps(schemas[tbl_id].jsonValue(), indent=4), "  "),
            )
            DeltaHandle.from_tc(tbl_id).drop_and_delete()
            affected_keys.append(tbl_id)



# Defines a class for opening a connection to Cosmos DB,
# loading and saving tables, and executing SQL.
import hashlib

# e.g. usage:
#   server = CosmosDb()
#   table_name = CosmosDb.table_name("TableId")
#   server.execute_sql(
#       f"CREATE TABLE cosmosCatalog.database_name.table_name using cosmos.oltp"
#       )
#   df = server.load_table("TableId")
#   server.save_table(df, "TableId")
from typing import Optional, Union

from azure.cosmos import CosmosClient, DatabaseProxy
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType

from spetlr.configurator.configurator import Configurator
from spetlr.cosmos.cosmos_base_server import CosmosBaseServer
from spetlr.cosmos.cosmos_handle import CosmosHandle
from spetlr.exceptions import NoSuchSchemaException, SpetlrException
from spetlr.schema_manager import SchemaManager
from spetlr.spark import Spark


class SpetlrCosmosException(SpetlrException):
    pass


class CosmosDb(CosmosBaseServer):
    def __init__(
        self,
        account_key: str,
        database: str,
        account_name: str = None,
        endpoint: str = None,
        catalog_name: Optional[str] = "cosmosCatalog",
    ):
        if not account_name and not endpoint:
            raise ValueError("account_name or endpoint must be set")

        if not endpoint:
            endpoint_pattern = "https://{}.documents.azure.com:443/"
            endpoint = endpoint_pattern.format(account_name)

        self.endpoint = endpoint
        self.account_key = account_key
        self.database = database
        self.config = {
            "spark.cosmos.accountEndpoint": self.endpoint,
            "spark.cosmos.accountkey": account_key,
            "spark.cosmos.database": database,
            "spark.cosmos.container": None,
        }
        self.client = CosmosClient(endpoint, credential=account_key)

        self._db_client: Optional[DatabaseProxy] = None
        self.catalog_name = (
            catalog_name
            or hashlib.sha1(f"{endpoint}{self.database}".encode()).hexdigest()
        )

    @property
    def db_client(self):
        if self._db_client is not None:
            return self._db_client

        self._db_client = self.client.get_database_client(self.database)
        return self._db_client

    def execute_sql(self, sql: str):
        # Examples:
        #
        # sql = f"CREATE DATABASE IF NOT EXISTS cosmosCatalog.{database_name};"
        #
        # sql = f"CREATE TABLE IF NOT EXISTS cosmosCatalog.{database_name}.{table_name}"
        #       " using cosmos.oltp "
        #       "TBLPROPERTIES(partitionKeyPath = '/id', manualThroughput = '1100')"
        spark = Spark.get()
        spark.conf.set(
            f"spark.sql.catalog.{self.catalog_name}",
            "com.azure.cosmos.spark.CosmosCatalog",
        )
        spark.conf.set(
            f"spark.sql.catalog.{self.catalog_name}.spark.cosmos.accountEndpoint",
            self.endpoint,
        )
        spark.conf.set(
            f"spark.sql.catalog.{self.catalog_name}.spark.cosmos.accountKey",
            self.account_key,
        )
        return spark.sql(sql)

    def read_table_by_name(self, table_name: str, schema: DataType = None) -> DataFrame:
        config = self.config.copy()
        config["spark.cosmos.container"] = table_name
        rd = Spark.get().read.format("cosmos.oltp").options(**config)
        if schema is not None:
            # noinspection PyTypeChecker
            rd = rd.schema(schema)
        else:
            rd = rd.option("spark.cosmos.read.inferSchema.enabled", "true")
        return rd.load()

    def read_table(self, table_id: str, schema: DataType = None) -> DataFrame:
        table_name = Configurator().table_name(table_id)
        return self.read_table_by_name(table_name, schema)

    def write_table_by_name(
        self, df_source: DataFrame, table_name: str, rows_per_partition: int = None
    ):
        if (
            rows_per_partition is not None
            and df_source.count() > rows_per_partition * 2
        ):
            partitions = int(1 + df_source.count() / rows_per_partition)
            df_source = df_source.repartition(partitions)
        config = self.config.copy()
        config["spark.cosmos.container"] = table_name
        (
            df_source.write.format("cosmos.oltp")
            .options(**config)
            .mode("append")  # overwrite is not supported in CosmosDB
            .save()
        )

    def write_table(
        self, df_source: DataFrame, table_id: str, rows_per_partition: int = None
    ):
        table_name = Configurator().table_name(table_id)
        self.write_table_by_name(df_source, table_name, rows_per_partition)

    def delete_item(
        self, table_id: str, id: Union[int, str], pk: Union[int, str] = None
    ):
        cntr = self.db_client.get_container_client(Configurator().table_name(table_id))
        cntr.delete_item(id, partition_key=pk)

    def delete_container(self, table_id: str):
        self.delete_container_by_name(Configurator().table_name(table_id))

    def delete_container_by_name(self, table_name: str):
        self.db_client.delete_container(table_name)

    def recreate_container_by_name(self, table_name: str):
        """
        Delete and recreate the container while preserving properties as
        far as possible.
        """

        for container in self.db_client.list_containers():
            if container["id"] == table_name:
                break
        else:
            raise SpetlrCosmosException(f"table not found {table_name}")

        throughput_units = (
            self.db_client.get_container_client(table_name)
            .get_throughput()
            .offer_throughput
        )

        self.db_client.delete_container(table_name)
        self.db_client.create_container(
            id=container["id"],
            partition_key=container["partitionKey"],
            offer_throughput=throughput_units,
            default_ttl=container.get("defaultTtl", None),
            indexing_policy=container.get("indexingPolicy", None),
        )

    def from_tc(self, table_id: str) -> CosmosHandle:
        tc = Configurator()
        name = tc.table_name(table_id)
        rows_per_partition = tc.table_property(table_id, "rows_per_partition", "")
        rows_per_partition = int(rows_per_partition) if rows_per_partition else None

        try:
            schema = SchemaManager().get_schema(table_id)
        except NoSuchSchemaException:
            schema = None

        return CosmosHandle(
            name=name,
            cosmos_db=self,
            schema=schema,
            rows_per_partition=rows_per_partition,
        )



# Defines input parameters for a cached Loader

from typing import List


class CachedLoaderParameters:
    def __init__(
        self,
        cache_table_name: str,
        key_cols: List[str],
        cache_id_cols: List[str] = None,
    ):
        """
        Args:
            cache_table_name: The table that holds the cache
            key_cols: the set of columns that form the primary key for a row
            cache_id_cols: These columns, added by the write operation, will be saved
                in the cache to identify e.g. the written batch.

        The table cache_table_name must exist and must have the following schema:
        (
            [key_cols column definitions],
            rowHash INTEGER,
            loadedTime TIMESTAMP,
            deletedTime TIMESTAMP,
            [cache_id_cols definitions, (if used)]
        )
        """

        self.cache_id_cols = cache_id_cols or []
        self.cache_table_name = cache_table_name
        self.key_cols = key_cols
        if not key_cols:
            raise ValueError("key columns must be provided")

        self.rowHash = "rowHash"
        self.loadedTime = "loadedTime"
        self.deletedTime = "deletedTime"



import sys
import unittest
from tempfile import NamedTemporaryFile
from textwrap import dedent

from spetlr import Configurator
from spetlr.exceptions.cli_exceptions import SpetlrCliCheckFailed

from . import tables1


class TestConfiguratorCli(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        c = Configurator()
        c.clear_all_configurations()

    def test_01_import_config(self):
        c = Configurator()
        c.add_resource_path(tables1)

        name = "configurator_cli_tmp_file"

        with NamedTemporaryFile() as nf:
            name = nf.name
            nf.close()
            sys.argv = ["mycliprog", "generate-keys-file", "-c", "-o", name]
            with self.assertRaises(SpetlrCliCheckFailed):
                # file did not exist. exit code 1
                c.cli()

            sys.argv = ["mycliprog", "generate-keys-file", "--output-file", name]
            c.cli()
            # file written. clean exit

            conts = open(name).read()
            expected = dedent(
                """\
                # AUTO GENERATED FILE
                # contains all spetlr.Configurator keys

                ID = "ID"
                MNT = "MNT"
                MyAlias = "MyAlias"
                MyFirst = "MyFirst"
                MyForked = "MyForked"
                MyRecursing = "MyRecursing"
                MySecond = "MySecond"
            """
            )
            self.assertEqual(conts, expected)

            # repeat the test
            sys.argv = ["mycliprog", "generate-keys-file", "--check-only", "-o", name]
            c.cli()
            # check passes since the file contents were ok

            with open(name, "w") as f:
                f.write(expected[:-10])  # bad contents

            sys.argv = ["mycliprog", "generate-keys-file", "-c", "-o", name]
            with self.assertRaises(SpetlrCliCheckFailed):
                # file had bad contents
                c.cli()



import pyspark.sql.functions as F
import pyspark.sql.types as T
from pyspark.sql import DataFrame


def _lower(col: str, caseInsensitiveMatching: bool):
    return col.lower() if caseInsensitiveMatching else col


def SelectAndCastColumns(
    *, df: DataFrame, schema: T.StructType, caseInsensitiveMatching: bool = False
) -> DataFrame:
    """
    Select and cast columns based pyspark schema.
    In case columns from shema are missing, they will be added with None values.

    param df: The pyspark DataFrame.
    param schema: pyspark schema to use for select and cast.
    param caseInsensitiveMatching: Boolean used for case insensitive matching,
    by default False.

    return: DataFrame mathing schema
    """

    dfColumns = [_lower(col, caseInsensitiveMatching) for col in df.columns]

    selectAndCastColumnsList = [
        F.col(c.name).cast(c.dataType).alias(c.name)
        if _lower(c.name, caseInsensitiveMatching) in dfColumns
        else F.lit(None).cast(c.dataType).alias(c.name)
        for c in schema
    ]

    return df.select(selectAndCastColumnsList)



import sqlparse
from sqlparse.sql import Statement


def _meaningful_token_iter(tks: Statement):
    """Only return tokens that are potentially meaningful sql"""
    for tk in tks:
        if (
            tk.ttype not in sqlparse.tokens.Whitespace
            and tk.ttype not in sqlparse.tokens.Comment
        ):
            yield tk



from pyspark.sql.types import IntegerType, StringType, StructField, StructType
from spetlrtools.testing import DataframeTestCase

from spetlr.spark import Spark
from spetlr.transformers.data_change_capture_transformer import (
    DataChangeCaptureTransformer,
)


class CaptureHistoryTransformerTest(DataframeTestCase):
    def setUp(self):
        self.transformer = DataChangeCaptureTransformer("id")
        self.schema = StructType(
            [
                StructField("id", IntegerType(), True),
                StructField("firstName", StringType(), True),
                StructField("lastName", StringType(), True),
                StructField("age", IntegerType(), True),
            ]
        )

        self.columns = ["id", "firstName", "lastName", "age"]

        row_1_1 = (1, "Peter", "Plys", 7)
        row_1_2 = (1, "Peter", "Plys", 8)
        row_2_1 = (2, "Peter", "Pan", 14)
        row_2_2 = (2, "Peter", "PanCake", 14)
        row_3_1 = (3, "Ole", "Lukoje", 98)

        # Test data test_01
        self.source1 = [row_1_1]
        self.target1 = []
        self.expected1 = self.source1

        # Test data for test_02
        self.source2 = [row_1_2, row_2_1]
        self.target2 = [row_1_1]
        self.expected2 = self.source2

        # Test data for test_03
        self.source3 = [row_1_2, row_2_2, row_3_1]
        self.target3 = [row_1_2, row_2_1]
        self.expected3 = [row_2_2, row_3_1]

        # Test data for test_04
        self.source4 = [row_1_2, row_2_2, row_3_1]
        self.target4 = [row_1_2, row_2_2, row_3_1]
        self.expected4 = []

        # Test data for test_05
        self.source5 = [row_1_1]
        self.target5 = [row_1_2]
        self.expected5 = []

    def test_01_capture_new_rows(self):
        """Target is empty, return all records from source"""

        df_source = Spark.get().createDataFrame(data=self.source1, schema=self.schema)
        df_target = Spark.get().createDataFrame(data=self.target1, schema=self.schema)

        dataset_group = {"source": df_source, "target": df_target}

        result_df = self.transformer.process_many(dataset_group)

        self.assertDataframeMatches(
            df=result_df, columns=self.columns, expected_data=self.expected1
        )

    def test_02_capture_new_and_modified_rows(self):
        """returns modified and new rows"""

        df_source = Spark.get().createDataFrame(data=self.source2, schema=self.schema)
        df_target = Spark.get().createDataFrame(data=self.target2, schema=self.schema)

        dataset_group = {"source": df_source, "target": df_target}

        result_df = self.transformer.process_many(dataset_group)

        self.assertDataframeMatches(
            df=result_df, columns=self.columns, expected_data=self.expected2
        )

    def test_03_capture_new_and_modified_rows_ignore_existing(self):
        """returns modified and new rows, and ignore existing, that is unchanged."""

        df_source = Spark.get().createDataFrame(data=self.source3, schema=self.schema)
        df_target = Spark.get().createDataFrame(data=self.target3, schema=self.schema)

        dataset_group = {"source": df_source, "target": df_target}

        result_df = self.transformer.process_many(dataset_group)

        self.assertDataframeMatches(
            df=result_df, columns=self.columns, expected_data=self.expected3
        )

    def test_04_no_changes(self):
        """returning dataframe is empty, since no changes have been detected."""

        df_source = Spark.get().createDataFrame(data=self.source4, schema=self.schema)
        df_target = Spark.get().createDataFrame(data=self.target4, schema=self.schema)

        dataset_group = {"source": df_source, "target": df_target}

        result_df = self.transformer.process_many(dataset_group)

        self.assertDataframeMatches(
            df=result_df, columns=self.columns, expected_data=self.expected4
        )

    def test_05_exclude_colc(self):
        """Ignore age-column, and therefore no data expected"""

        df_source = Spark.get().createDataFrame(data=self.source5, schema=self.schema)
        df_target = Spark.get().createDataFrame(data=self.target5, schema=self.schema)

        dataset_group = {"source": df_source, "target": df_target}

        result_df = DataChangeCaptureTransformer(
            primary_key="id", cols_to_exclude=["age"]
        ).process_many(dataset_group)

        self.assertDataframeMatches(
            df=result_df, columns=self.columns, expected_data=self.expected5
        )



import unittest

import pyspark.sql.functions as f
from pyspark.sql.types import StringType, StructField, StructType

import spetlr.transformations as spetlr_transform
from spetlr.spark import Spark


class ConcatDfTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.df1 = create_df1()
        cls.df2 = create_df2()
        cls.df3 = create_df3()

    def test_concat_one_df(self):
        result = spetlr_transform.concat_dfs([self.df1])

        # Test that the columns are there
        cols = result.columns
        self.assertIn("id", cols)
        self.assertIn("brand", cols)
        self.assertIn("model", cols)
        self.assertIn("year", cols)

        # Test the number of rows
        self.assertEqual(3, result.count())
        self.assertEqual(
            1, get_number_rows_1(result, "1", "Fender", "Telecaster", year="1950")
        )
        self.assertEqual(
            1, get_number_rows_1(result, "2", "Gibson", "Les Paul", year="1959")
        )
        self.assertEqual(1, get_number_rows_1(result, "3", "Ibanez", "RG", year="1987"))

        # Test that there is only one of each column
        self.assertEqual(1, cols.count("id"))
        self.assertEqual(1, cols.count("brand"))
        self.assertEqual(1, cols.count("model"))
        self.assertEqual(1, cols.count("year"))

    def test_concat_two_df(self):
        result = spetlr_transform.concat_dfs([self.df1, self.df2])

        # Test that the columns are there
        cols = result.columns
        self.assertIn("id", cols)
        self.assertIn("brand", cols)
        self.assertIn("model", cols)
        self.assertIn("year", cols)
        self.assertIn("size", cols)

        # Test the number of rows
        self.assertEqual(6, result.count())
        self.assertEqual(
            1, get_number_rows_2(result, "1", "Fender", "Telecaster", year="1950")
        )
        self.assertEqual(
            1, get_number_rows_2(result, "2", "Gibson", "Les Paul", year="1959")
        )
        self.assertEqual(1, get_number_rows_2(result, "3", "Ibanez", "RG", year="1987"))
        self.assertEqual(
            1, get_number_rows_2(result, "1", "Fender", "Stratocaster", size="Small")
        )
        self.assertEqual(
            1,
            get_number_rows_2(result, "2", "Gibson", "Les Paul Junior", size="Medium"),
        )
        self.assertEqual(
            1, get_number_rows_2(result, "3", "Ibanez", "JPM", size="Large")
        )

        # Test that there is only one of each column
        self.assertEqual(1, cols.count("id"))
        self.assertEqual(1, cols.count("brand"))
        self.assertEqual(1, cols.count("model"))
        self.assertEqual(1, cols.count("year"))
        self.assertEqual(1, cols.count("size"))

    def test_concat_three_df(self):
        result = spetlr_transform.concat_dfs([self.df1, self.df2, self.df3])

        # Test that the columns are there
        cols = result.columns
        self.assertIn("id", cols)
        self.assertIn("brand", cols)
        self.assertIn("model", cols)
        self.assertIn("year", cols)
        self.assertIn("size", cols)
        self.assertIn("color", cols)

        # Test the number of rows
        self.assertEqual(9, result.count())
        self.assertEqual(
            1, get_number_rows_3(result, "1", "Fender", "Telecaster", year="1950")
        )
        self.assertEqual(
            1, get_number_rows_3(result, "2", "Gibson", "Les Paul", year="1959")
        )
        self.assertEqual(1, get_number_rows_3(result, "3", "Ibanez", "RG", year="1987"))
        self.assertEqual(
            1, get_number_rows_3(result, "1", "Fender", "Stratocaster", size="Small")
        )
        self.assertEqual(
            1,
            get_number_rows_3(result, "2", "Gibson", "Les Paul Junior", size="Medium"),
        )
        self.assertEqual(
            1, get_number_rows_3(result, "3", "Ibanez", "JPM", size="Large")
        )
        self.assertEqual(
            1, get_number_rows_3(result, "1", "Fender", "Jaguar", color="Brown")
        )
        self.assertEqual(
            1, get_number_rows_3(result, "2", "Gibson", "EB", color="Blue")
        )
        self.assertEqual(
            1, get_number_rows_3(result, "3", "Ibanez", "AE", color="Black")
        )

        # Test that there is only one of each column
        self.assertEqual(1, cols.count("id"))
        self.assertEqual(1, cols.count("brand"))
        self.assertEqual(1, cols.count("model"))
        self.assertEqual(1, cols.count("year"))
        self.assertEqual(1, cols.count("size"))
        self.assertEqual(1, cols.count("color"))


def create_df1():
    return Spark.get().createDataFrame(
        Spark.get().sparkContext.parallelize(
            [
                ("1", "Fender", "Telecaster", "1950"),
                ("2", "Gibson", "Les Paul", "1959"),
                ("3", "Ibanez", "RG", "1987"),
            ]
        ),
        StructType(
            [
                StructField("id", StringType()),
                StructField("brand", StringType()),
                StructField("model", StringType()),
                StructField("year", StringType()),
            ]
        ),
    )


def create_df2():
    return Spark.get().createDataFrame(
        Spark.get().sparkContext.parallelize(
            [
                ("1", "Fender", "Stratocaster", "Small"),
                ("2", "Gibson", "Les Paul Junior", "Medium"),
                ("3", "Ibanez", "JPM", "Large"),
            ]
        ),
        StructType(
            [
                StructField("id", StringType()),
                StructField("brand", StringType()),
                StructField("model", StringType()),
                StructField("size", StringType()),
            ]
        ),
    )


def create_df3():
    return Spark.get().createDataFrame(
        Spark.get().sparkContext.parallelize(
            [
                ("1", "Fender", "Jaguar", "Brown"),
                ("2", "Gibson", "EB", "Blue"),
                ("3", "Ibanez", "AE", "Black"),
            ]
        ),
        StructType(
            [
                StructField("id", StringType()),
                StructField("brand", StringType()),
                StructField("model", StringType()),
                StructField("color", StringType()),
            ]
        ),
    )


def get_number_rows_1(df, id, brand, model, year=None):
    # When testing the first transformation, theres only "year" as column
    if year is not None:
        return df.filter(
            (f.col("id") == id)
            & (f.col("brand") == brand)
            & (f.col("model") == model)
            & (f.col("year") == year)
        ).count()
    return None


def get_number_rows_2(df, id, brand, model, year=None, size=None):
    # When testing the second transformation, theres only "year" and "size" as column
    if year is not None:
        return df.filter(
            (f.col("id") == id)
            & (f.col("brand") == brand)
            & (f.col("model") == model)
            & (f.col("year") == year)
            & (f.col("size").isNull())
        ).count()
    if size is not None:
        return df.filter(
            (f.col("id") == id)
            & (f.col("brand") == brand)
            & (f.col("model") == model)
            & (f.col("year").isNull())
            & (f.col("size") == size)
        ).count()

    return None


def get_number_rows_3(df, id, brand, model, year=None, size=None, color=None):
    if year is not None:
        return df.filter(
            (f.col("id") == id)
            & (f.col("brand") == brand)
            & (f.col("model") == model)
            & (f.col("year") == year)
            & (f.col("size").isNull())
            & (f.col("color").isNull())
        ).count()
    if size is not None:
        return df.filter(
            (f.col("id") == id)
            & (f.col("brand") == brand)
            & (f.col("model") == model)
            & (f.col("year").isNull())
            & (f.col("size") == size)
            & (f.col("color").isNull())
        ).count()
    if color is not None:
        return df.filter(
            (f.col("id") == id)
            & (f.col("brand") == brand)
            & (f.col("model") == model)
            & (f.col("year").isNull())
            & (f.col("size").isNull())
            & (f.col("color") == color)
        ).count()

    return None



from .load_modes import Appendable, Overwritable, Upsertable  # noqa: F401
from .simple_loader import SimpleLoader  # noqa: F401
from .simple_sql_loader import SimpleSqlServerLoader  # noqa: F401
from .upsert_loader_streaming import UpsertLoaderStreaming  # noqa: F401
from .UpsertLoader import UpsertLoader  # noqa: F401



from unittest import TestCase

from spetlr.deltaspec.exceptions import InvalidTableName
from spetlr.deltaspec.helpers import TableName


class TestTableName(TestCase):
    def test_01_levels(self):
        n = TableName.from_str(None)
        self.assertEqual(n.level(), 0)
        self.assertEqual(str(n), "")
        self.assertEqual(TableName.from_str(None), TableName())
        self.assertEqual(TableName.from_str(""), TableName())

        t = TableName.from_str("tbl")
        self.assertEqual(t.level(), 1)
        self.assertEqual(str(t), "tbl")

        d = TableName.from_str("db.tbl")
        self.assertEqual(d.level(), 2)
        self.assertEqual(str(d), "db.tbl")

        c = TableName.from_str("ctlg.db.tbl")
        self.assertEqual(c.level(), 3)
        self.assertEqual(str(c), "ctlg.db.tbl")

        with self.assertRaises(InvalidTableName):
            TableName(schema="db")

        with self.assertRaises(InvalidTableName):
            TableName(catalog="c", schema="db")

        self.assertEqual(str(c.to_level(2)), "db.tbl")
        self.assertEqual(str(c.to_level(1)), "tbl")
        self.assertEqual(str(c.to_level(0)), "")



import dataclasses
import unittest
from textwrap import dedent

from pyspark.sql import types as t

from spetlr import Configurator
from spetlr.deltaspec import DeltaTableSpecDifference, TableSpectNotEnforcable
from spetlr.deltaspec.DeltaTableSpec import DeltaTableSpec
from tests.cluster.delta.deltaspec import tables
from tests.local.delta import sql


class TestDeltaTableSpec(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        c = Configurator()
        c.clear_all_configurations()
        c.set_prod()

        cls.base = tables.base
        cls.target = tables.target

    def test_01_create_statements(self):
        self.assertEqual(
            self.base.get_sql_create(),
            dedent(
                """\
                CREATE TABLE myDeltaTableSpecTestDb{ID}.tbl
                (
                  c double,
                  d string COMMENT "Whatsupp",
                  onlyb int,
                  a int,
                  b int
                )
                USING DELTA
                LOCATION "dbfs:/tmp/somewhere{ID}/over/the/rainbow"
                TBLPROPERTIES (
                  "delta.columnMapping.mode" = "name",
                  "delta.minReaderVersion" = "2",
                  "delta.minWriterVersion" = "5"
                )
                """
            ),
        )

    def test_02_alter_statements_raise(self):
        Configurator().set_prod()
        forward_diff = self.target.compare_to(self.base)
        with self.assertRaises(TableSpectNotEnforcable):
            forward_diff.alter_statements(
                allow_columns_add=False,
                allow_columns_type_change=True,
                allow_columns_drop=True,
                allow_columns_reorder=True,
            )

        with self.assertRaises(TableSpectNotEnforcable):
            forward_diff.alter_statements(
                allow_columns_add=True,
                allow_columns_type_change=False,
                allow_columns_drop=True,
                allow_columns_reorder=True,
            )

        with self.assertRaises(TableSpectNotEnforcable):
            forward_diff.alter_statements(
                allow_columns_add=True,
                allow_columns_type_change=True,
                allow_columns_drop=True,
                allow_columns_reorder=False,
            )
        print("No raise when they are all warnings")
        forward_diff.alter_statements(errors_as_warnings=True)

    def test_03_diff_alter_statements(self):
        Configurator().set_prod()
        forward_diff = self.target.compare_to(self.base)

        self.assertEqual(
            forward_diff.alter_statements(
                allow_columns_add=True,
                allow_columns_type_change=True,
                allow_columns_drop=True,
                allow_columns_reorder=True,
            ),
            [
                "ALTER TABLE mydeltatablespectestdb.tbl "
                'SET TBLPROPERTIES ("my.cool.peoperty" = "bacon")',
                "ALTER TABLE mydeltatablespectestdb.tbl DROP COLUMNS (b, onlyb)",
                dedent(
                    """\
                ALTER TABLE mydeltatablespectestdb.tbl ADD COLUMNS (
                  b string,
                  onlyt string COMMENT "Only in target"
                )"""
                ),
                # "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a DROP NOT NULL",
                # "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d SET NOT NULL",
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a COMMENT"
                ' "gains not null"',
                'ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d COMMENT ""',
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a FIRST",
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN b AFTER a",
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN onlyt AFTER d",
                'COMMENT ON TABLE mydeltatablespectestdb.tbl is "Contains useful data"',
            ],
        )

        reverse_diff = self.base.compare_to(self.target)
        self.assertEqual(
            reverse_diff.alter_statements(
                allow_columns_add=True,
                allow_columns_type_change=True,
                allow_columns_drop=True,
                allow_columns_reorder=True,
                allow_name_change=True,
                allow_location_change=True,
            ),
            [
                "ALTER TABLE mydeltatablespectestdb.tbl "
                'UNSET TBLPROPERTIES ("my.cool.peoperty")',
                "ALTER TABLE mydeltatablespectestdb.tbl DROP COLUMNS (b, onlyt)",
                dedent(
                    """\
                ALTER TABLE mydeltatablespectestdb.tbl ADD COLUMNS (
                  onlyb int,
                  b int
                )"""
                ),
                # "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d DROP NOT NULL",
                # "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a SET NOT NULL",
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d COMMENT "
                '"Whatsupp"',
                'ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a COMMENT ""',
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN c FIRST",
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d AFTER c",
                "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN onlyb AFTER d",
                "COMMENT ON TABLE mydeltatablespectestdb.tbl is null",
            ],
        )

    def test_04_check_init_protections(self):
        tbl1 = DeltaTableSpec(
            name="myDeltaTableSpecTestDb{ID}.tbl",
            schema=t.StructType(
                fields=[
                    t.StructField(name="c", dataType=t.DoubleType()),
                    t.StructField(
                        name="d",
                        dataType=t.StringType(),
                        nullable=False,
                        metadata={"comment": "Whatsupp"},
                    ),
                    t.StructField(name="onlyb", dataType=t.IntegerType()),
                    t.StructField(name="a", dataType=t.IntegerType()),
                    t.StructField(name="b", dataType=t.StringType()),
                ]
            ),
            location="/somewhere/over/the{ID}/rainbow",
        )
        tbl2 = DeltaTableSpec(
            name="spark_catalog.mydeltatablespectestdb.tbl",
            schema=t.StructType(
                fields=[
                    t.StructField(name="c", dataType=t.DoubleType()),
                    t.StructField(
                        name="d",
                        dataType=t.StringType(),
                        nullable=False,
                        metadata={"comment": "Whatsupp"},
                    ),
                    t.StructField(name="onlyb", dataType=t.IntegerType()),
                    t.StructField(name="a", dataType=t.IntegerType()),
                    t.StructField(name="b", dataType=t.StringType()),
                ]
            ),
            location="dbfs:/somewhere/over/the/rainbow",
        )
        Configurator().set_prod()
        d = tbl2.compare_to(tbl1.fully_substituted())
        self.assertFalse(d.is_different(), d)

    def test_05_namechagne(self):
        statements = tables.newname.compare_to(tables.oldname).alter_statements(
            allow_columns_add=True,
            allow_columns_type_change=True,
            allow_columns_drop=True,
            allow_columns_reorder=True,
            allow_name_change=True,
            allow_location_change=True,
        )
        self.assertEqual(
            statements,
            [
                "ALTER TABLE mydeltatablespectestdb.namechange_old RENAME TO "
                "mydeltatablespectestdb.namechange_new"
            ],
        )

    def test_06_location_change(self):
        statements = tables.newlocation.compare_to(tables.oldlocation).alter_statements(
            allow_columns_add=True,
            allow_columns_type_change=True,
            allow_columns_drop=True,
            allow_columns_reorder=True,
            allow_name_change=True,
            allow_location_change=True,
        )
        self.assertEqual(
            statements,
            [
                "CREATE TABLE delta.`dbfs:/tmp/somewhere/locchange/new`\n"
                "(\n"
                "  b string,\n"
                "  c double,\n"
                "  d string\n"
                ")\n"
                "USING DELTA\n"
                'LOCATION "dbfs:/tmp/somewhere/locchange/new"\n'
                'COMMENT "Contains useful data"\n'
                "TBLPROPERTIES (\n"
                '  "delta.columnMapping.mode" = "name",\n'
                '  "delta.minReaderVersion" = "2",\n'
                '  "delta.minWriterVersion" = "5"\n'
                ")\n",
                "ALTER TABLE mydeltatablespectestdb.locchange SET LOCATION "
                '"dbfs:/tmp/somewhere/locchange/new"',
            ],
        )

    def test_07_from_tc(self):
        c = Configurator()
        c.clear_all_configurations()
        c.add_sql_resource_path(sql)

        ds = DeltaTableSpec.from_tc("Table1")
        self.assertEqual(
            ds.get_sql_create(),
            (
                "CREATE TABLE sometable\n"
                "(\n"
                "  a int\n"
                ")\n"
                "USING DELTA\n"
                "TBLPROPERTIES (\n"
                '  "delta.columnMapping.mode" = "name",\n'
                '  "delta.minReaderVersion" = "2",\n'
                '  "delta.minWriterVersion" = "5"\n'
                ")\n"
            ),
        )

        # The deltaspec needs to be resolved explicitly
        ds = DeltaTableSpec.from_tc("Table2")
        c.set_prod()
        self.assertEqual(ds.fully_substituted().location, "dbfs:/mnt/foo/bar")
        # This allows it to be globally initialized before the prod/debug is set
        c.set_debug()
        self.assertEqual(ds.fully_substituted().location, "dbfs:/tmp/foo/bar")

    def test_08_modify_to_allow_drop(self):
        Configurator().set_prod()
        # a table has been created without the DeltaTableSpec:
        base = DeltaTableSpec(
            name="myDeltaTableSpecTestDb{ID}.direct",
            schema=t.StructType(
                fields=[
                    t.StructField(name="b", dataType=t.StringType()),
                    t.StructField(name="c", dataType=t.DoubleType()),
                    t.StructField(name="d", dataType=t.StringType()),
                ]
            ),
            tblproperties={
                "delta.minReaderVersion": "2",
                "delta.minWriterVersion": "3",
            },
        )

        # now a modified version is to be applied that drops columns:
        ds = DeltaTableSpec.from_sql(tables.simple_modified_sql)
        diff = ds.compare_to(base)
        self.assertEqual(
            diff.alter_statements(allow_columns_drop=True),
            [
                "ALTER TABLE mydeltatablespectestdb.direct SET TBLPROPERTIES "
                '("delta.minWriterVersion" = "5", "delta.columnMapping.mode" = "name")',
                "ALTER TABLE mydeltatablespectestdb.direct DROP COLUMN (d)",
            ],
        )

    def test_09_generated_and_default_expression(self):
        c = Configurator()
        c.clear_all_configurations()
        c.add_sql_resource_path(sql)
        ds = DeltaTableSpec.from_tc("Table3")
        self.assertEqual(
            ds.schema,
            t.StructType(
                [
                    t.StructField("id", t.LongType(), True),
                    t.StructField("a", t.IntegerType(), True),
                    t.StructField("b", t.IntegerType(), True),
                    t.StructField("area", t.IntegerType(), True),
                ]
            ),
        )
        diff = DeltaTableSpecDifference(
            target=ds, base=dataclasses.replace(ds, tblproperties={})
        )
        # The diff is not a match, tableproperties, mismatch
        self.assertFalse(diff.complete_match(), diff)

        # ignore the tblproperties and it matches
        diff = diff.ignore("tblproperties")
        self.assertTrue(diff.complete_match(), diff)



import datetime

import pyspark.sql.functions as f
from pyspark.sql import DataFrame

from spetlr.etl import Transformer
from spetlr.tables import TableHandle


class EhToDeltaBronzeTransformer(Transformer):
    """
    This class transforms eventhub data into the following schema:

    | EventhubRowId | BodyId | Body | EnqueuedTimestamp | StreamingTime |->
    |---------------|--------|------|-------------------|---------------|->
    |...            |...     |...   |...                |...            |->

    | SequenceNumber | Offset | SystemProperties | Properties| pdate |
    |----------------|--------|------------------|-----------|-------|
    |...             |...     |...               |...        |...    |

    Parameters:
    target_dh: DeltaHandle for the target delta table (bronze)
    df: A dataframe containing raw eventhub data

    Returns:
    A dataframe with the above mentioned schema


    """

    def __init__(self, target_dh: TableHandle):
        super().__init__()
        self.target_dh = target_dh
        self._eh_cols = [
            "EventhubRowId",
            "BodyId",
            "Body",
            "EnqueuedTimestamp",
            "StreamingTime",
            "SequenceNumber",
            "Offset",
            "SystemProperties",
            "Properties",
            "pdate",
        ]

    def process(self, df: DataFrame) -> DataFrame:
        target_df = self.target_dh.read()

        assert set(self._eh_cols).issubset(target_df.columns)

        # Generate Unique id for the eventhub rows
        df = df.withColumn(
            "EventhubRowId",
            f.conv(
                f.concat_ws(
                    "",
                    f.lit("0"),
                    f.substring(
                        f.concat_ws(
                            "",
                            f.sha2(f.col("Body").cast("string"), 256),
                            f.sha2(f.col("EnqueuedTimestamp").cast("string"), 256),
                        ),
                        -15,
                        15,
                    ),
                ),
                16,
                10,
            ).cast("long"),
        )

        # Generate id for the eventhub rows using hashed body
        # Can be used for identify rows with same body
        df = df.withColumn(
            "BodyId",
            f.conv(
                f.concat_ws(
                    "",
                    f.lit("0"),
                    f.substring(f.sha2(f.col("Body").cast("string"), 256), -15, 15),
                ),
                16,
                10,
            ).cast("long"),
        )

        # Add streaming time
        streaming_time = datetime.datetime.now(datetime.timezone.utc).replace(
            microsecond=0
        )
        df = df.withColumn("StreamingTime", f.lit(streaming_time))

        # Cast body to string
        df = df.select(
            f.col("EventhubRowId").cast("long").alias("EventhubRowId"),
            f.col("BodyId").cast("long").alias("BodyId"),
            f.col("Body").cast("string").alias("Body"),
            f.col("EnqueuedTimestamp").cast("timestamp").alias("EnqueuedTimestamp"),
            f.col("StreamingTime").cast("timestamp").alias("StreamingTime"),
            f.col("SequenceNumber").cast("long").alias("SequenceNumber"),
            f.col("Offset").cast("string").alias("Offset"),
            f.col("SystemProperties").cast("string").alias("SystemProperties"),
            f.col("Properties").cast("string").alias("Properties"),
            f.col("pdate").cast("timestamp").alias("pdate"),
        )

        # Ensure that cols are selected correctly
        df = df.select(*self._eh_cols)

        return df



"""
This script installs and runs git commit hooks.
To install the hooks, please run

$> python git_hooks.py

To uninstall the git hooks, run

$> python git_hooks.py uninstall

"""
import os
import subprocess
import sys
from pathlib import Path
from textwrap import dedent

actions = {}


def action(func):
    global actions
    actions[func.__name__.replace("_", "-")] = func


@action
def pre_commit():
    """
    This pre-commit hook runs the black code formatter on all changed files
    that end in ".py"
    """
    print("spetlr pre-commit hook")

    # this command
    # $> git diff --cached --name-only --diff-filter=d
    # results in an output like
    # src/spetlr/formatting/git_hooks.py
    # with one line per Modified or New file, deleted files are excluded
    # we don't want to reformat deleted files, all others should be formatted.
    files_to_check = []
    for line in (
        subprocess.run(
            [
                "git",
                "diff",
                "--cached",
                # only show the names, one name per line
                "--name-only",
                # diff filter lowercase d means exclude deleted files
                "--diff-filter=d",
            ],
            capture_output=True,
        )
        .stdout.decode()
        .splitlines()
    ):
        path = line.strip()
        if not path.endswith(".py"):
            continue
        files_to_check.append(path)

    if not files_to_check:
        print("  Nothing to do.")
        return

    # first reformat all affected python files
    subprocess.run(["black", *files_to_check], check=True)
    # now add all reformatted files back to be committed
    subprocess.run(["git", "add", *files_to_check], check=True)


def get_hooks_dir() -> Path:
    for path in [Path.cwd()] + list(Path().cwd().parents):
        if (path / ".git").is_dir():
            break
    else:
        raise AssertionError(
            "Git hooks directory not found. Please run this script form the repo base."
        )

    return path / ".git" / "hooks"


def install() -> None:
    print("Now installing hooks.")

    hooks_dir = get_hooks_dir()

    for command in actions.keys():
        with open(hooks_dir / command, "w") as f:
            print(f"Installing hook for '{command}'")
            f.write(
                dedent(
                    rf"""
    #!{sys.executable}
    import sys, subprocess
    if sys.executable != {repr(sys.executable)}:
        print("Changing interpreter")
        subprocess.run([{repr(sys.executable)}, __file__]+sys.argv[1:],check=True)
        sys.exit(0)

    try:
        from spetlr.formatting.git_hooks import actions
        actions[{repr(command)}]()
    except ModuleNotFoundError as e:
        import sys
        print(e)
        print('search path was:\n-', '\n- '.join(sys.path))
        print('interpreter:', sys.executable)

    """
                ).strip()
            )
    print("Done installing hooks.")


def uninstall() -> None:
    print("uninstalling git hooks")
    hooks_dir = get_hooks_dir()
    for command in actions.keys():
        print(f"  uninstalling '{command}'")
        os.unlink(hooks_dir / command)


def main():
    if "uninstall" in sys.argv:
        uninstall()
    else:
        install()



import json
from datetime import datetime

from pyspark.sql import DataFrame

from spetlr.etl.extractor import Extractor
from spetlr.spark import Spark


class InvalidEventhubStreamExtractorParameters(Exception):
    pass


class EventhubStreamExtractor(Extractor):
    def __init__(
        self,
        consumerGroup: str,
        connectionString: str = None,
        namespace: str = None,
        eventhub: str = None,
        accessKeyName: str = None,
        accessKey: str = None,
        maxEventsPerTrigger: int = 10000,
        startEnqueuedTime: datetime = None,
    ):
        """
        :param consumerGroup: the eventhub consumerGroup to use for streaming
        :param connectionString: connectionString to the eventhub,
            if not supplied namespace, eventhub,
            accessKeyName and accessKey have to be instead
        :param namespace: the eventhub namespace to use for
            streaming, can be ignored if connectionString is supplied
        :param eventhub: the eventhub name to use for streaming,
            can be ignored if connectionString is supplied
        :param accessKeyName: the eventhub accessKeyName to use for
            streaming, can be ignored if connectionString is supplied
        :param accessKey: the eventhub accessKey to use for streaming,
            can be ignored if connectionString is supplied
        :param maxEventsPerTrigger: the number of events handled
            per mico trigger in stream
        :param startEnqueuedTime: timestamp to define where stream starts,
            if None the stream wil start from the oldest event in eventhub
        """

        self.spark = Spark.get()

        # If connectionString is missing,
        # create it from namespace, eventhub, accessKeyName and accessKey
        # Raise exception is parameters are missing
        if connectionString is None:
            if (
                namespace is None
                or eventhub is None
                or accessKeyName is None
                or accessKey is None
            ):
                raise InvalidEventhubStreamExtractorParameters(
                    "Either connectionString or "
                    "(namespace, eventhub, accessKeyName and accessKey) "
                    "have to be supplied"
                )

            self.connectionString = (
                f"Endpoint=sb://{namespace}.servicebus.windows.net/{eventhub};"
                f"EntityPath={eventhub};SharedAccessKeyName={accessKeyName};"
                f"SharedAccessKey={accessKey}"
            )
        else:
            self.connectionString = connectionString

        self.consumerGroup = consumerGroup
        self.maxEventsPerTrigger = maxEventsPerTrigger

        # Define where to start eventhub stream
        # It can be done from offset, seqence number or timestamp
        self.startingEventPosition = {
            "offset": None,  # not in use
            "seqNo": -1,  # not in use
            "enqueuedTime": None,  # not in use
            "isInclusive": True,
        }

        if startEnqueuedTime:
            self.startingEventPosition["enqueuedTime"] = startEnqueuedTime.strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ"
            )  # Start from timestamp
        else:
            # Default is to start from beginning of stream
            self.startingEventPosition["offset"] = "-1"  # Start stream from beginning

    def read(self) -> DataFrame:
        print("Read eventhub data stream")

        config = {
            "eventhubs.connectionString": (
                self.spark.sparkContext._jvm.org.apache.spark.eventhubs.EventHubsUtils.encrypt(  # noqa: E501
                    self.connectionString
                )
            ),
            "maxEventsPerTrigger": self.maxEventsPerTrigger,
            "eventhubs.consumerGroup": self.consumerGroup,
            "eventhubs.startingPosition": json.dumps(self.startingEventPosition),
        }

        df = self.spark.readStream.format("eventhubs").options(**config).load()

        return df



from typing import List

import pyspark.sql.functions as F
from pyspark.sql import DataFrame
from timezonefinder import TimezoneFinder

from spetlr.etl import Transformer
from spetlr.etl.types import dataset_group
from spetlr.exceptions import ColumnDoesNotExistException


class TimeZoneTransformer(Transformer):
    """
    This transformer extracts a timezone using longitude and latitude.

    Attributes:
    ----------
        latitude_col : str
            name of column with latitude values
        longitude_col : str
            name of column with longitude values
        dataset_input_key : str
            input dataset identifier
        dataset_output_key : str
            output dataset identifier
        column_output_name : str
            output column name of the TimeZone column

    Methods
    -------
    process(df: DataFrame):
        returns the input DataFrame with a TimeZone column
    """

    def __init__(
        self,
        *,
        latitude_col: str,
        longitude_col: str,
        column_output_name: str = "TimeZone",
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.latitude_col = latitude_col
        self.longitude_col = longitude_col
        self.column_output_name = column_output_name

    def process(self, df: DataFrame) -> DataFrame:
        """
        Extracts a timezone using latitude and longitude columns,
        and outputs the input DataFrame with a TimeZone column.
        """
        columns = df.columns

        if self.latitude_col not in columns:
            raise ColumnDoesNotExistException(
                "The specified latitude column is not in the DataFrame"
            )

        if self.longitude_col not in columns:
            raise ColumnDoesNotExistException(
                "The specified longitude column is not in the DataFrame"
            )

        timezone_extractor = F.udf(
            lambda latitude, longitude: None
            if latitude is None or longitude is None
            else TimezoneFinder().timezone_at(lat=latitude, lng=longitude)
        )

        df = df.withColumn(
            self.column_output_name,
            timezone_extractor(F.col(self.latitude_col), F.col(self.longitude_col)),
        )

        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        raise NotImplementedError()



import uuid as _uuid

import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.configurator import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.sql import SqlExecutor
from spetlr.utils import DeleteMismatchedSchemas
from tests.cluster.utils import extras


class DeleteSchemaMismatch(DataframeTestCase):

    """
    Test class for DeleteMismatchedSchemas.

    This class contains tests for the `DeleteMismatchedSchemas` utility class.
    It verifies the behavior of the utility when dealing with Delta tables
    that have schema mismatches.

    The class defines the initial and changed schema for testing purposes.
    It also sets up the necessary configuration and resources for the tests.

    NB: This test is special, since it expects production tables.
        Do not use .set_prod() in tests.

    """

    initial_schema_delta = T.StructType(
        [
            T.StructField("a", T.IntegerType(), True),
        ]
    )

    changed_schema_delta = T.StructType(
        [
            T.StructField("a", T.IntegerType(), True),
            T.StructField("b", T.IntegerType(), True),
        ]
    )

    @classmethod
    def setUpClass(cls) -> None:
        Configurator().register("mismatchuuid", str(_uuid.uuid4().hex))
        Configurator().add_resource_path(extras)

        # This extra configuration is necessary
        # since SPETLR tests runs parallel on multiple cluster
        # but in same Databricks workspace
        # The production tables will interfere with each other without an unique id

        # This is a special case.
        # Dont use set_prod in testcases
        Configurator().set_prod()

        cls.executor = SqlExecutor(base_module=extras)

    def test_01_no_change_delta(self):
        """Test scenario with no schema change in Delta table."""
        dbhandle = DbHandle.from_tc("SparkTestDbMismatch")
        dbhandle.drop_cascade()

        # Setup production table

        self.executor.execute_sql_file("initial_schema")

        dh = DeltaHandle.from_tc("MismatchSparkTestTable")

        self.assertEqual(dh.read().schema, self.initial_schema_delta)

        DeleteMismatchedSchemas(
            spark_executor=self.executor, sql_files_pattern="initial_schema"
        )

        self.assertEqual(dh.read().schema, self.initial_schema_delta)

    def test_02_change_delta(self):
        """Test scenario with schema change in Delta table."""
        # Ensure that the database is dropped
        dbhandle = DbHandle.from_tc("SparkTestDbMismatch")
        dbhandle.drop_cascade()

        # Setup production table
        self.executor.execute_sql_file("initial_schema")

        dh = DeltaHandle.from_tc("MismatchSparkTestTable")

        self.assertEqual(dh.read().schema, self.initial_schema_delta)

        DeleteMismatchedSchemas(
            spark_executor=self.executor, sql_files_pattern="changed_schema"
        )
        self.executor.execute_sql_file("changed_schema")

        self.assertEqual(dh.read().schema, self.changed_schema_delta)



import hashlib

import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.spark import Spark
from spetlr.transformers import GenerateMd5ColumnTransformer


class TestGenerateMd5ColumnTransformer(DataframeTestCase):
    def test_generate_md5_column(self):
        input_schema = T.StructType(
            [
                T.StructField("id", T.IntegerType(), True),
                T.StructField("text", T.StringType(), True),
            ]
        )

        input_data = [
            (1, "text1"),
            (2, "text2"),
        ]

        df_input = Spark.get().createDataFrame(data=input_data, schema=input_schema)

        df_transformed = GenerateMd5ColumnTransformer(
            col_name="md5_col",
            col_list=["id", "text"],
        ).process(df_input)

        expected_data = [
            (1, "text1", hashlib.md5("1text1".encode()).hexdigest()),
            (2, "text2", hashlib.md5("2text2".encode()).hexdigest()),
        ]

        self.assertDataframeMatches(
            df=df_transformed,
            expected_data=expected_data,
        )

    def test_generate_md5_column_with_nulls(self):
        input_schema = T.StructType(
            [
                T.StructField("id", T.IntegerType(), True),
                T.StructField("text", T.StringType(), True),
            ]
        )

        input_data = [
            (1, "text1"),
            (2, None),
        ]

        df_input = Spark.get().createDataFrame(data=input_data, schema=input_schema)

        df_transformed = GenerateMd5ColumnTransformer(
            col_name="md5_col",
            col_list=["id", "text"],
        ).process(df_input)

        expected_data = [
            (1, "text1", hashlib.md5("1text1".encode()).hexdigest()),
            (2, None, hashlib.md5(("2" + "").encode()).hexdigest()),
        ]

        self.assertDataframeMatches(
            df=df_transformed,
            expected_data=expected_data,
        )



from .DataframeCreator import DataframeCreator
from .DeleteMismatchedSchemas import DeleteMismatchedSchemas
from .DropOldestDuplicates import DropOldestDuplicates
from .GetMergeStatement import GetMergeStatement
from .MockExtractor import MockExtractor
from .MockLoader import MockLoader
from .SelectAndCastColumns import SelectAndCastColumns

__all__ = [
    DataframeCreator,
    MockLoader,
    GetMergeStatement,
    MockExtractor,
    DropOldestDuplicates,
    SelectAndCastColumns,
    DeleteMismatchedSchemas,
]



import warnings
from unittest.mock import Mock

from spetlrtools.testing import DataframeTestCase

from spetlr.etl.loaders.stream_loader import StreamLoader
from spetlr.exceptions import (
    InvalidStreamTriggerType,
    NeedTriggerTimeWhenProcessingType,
    UnknownStreamOutputMode,
)


class StreamLoaderTests(DataframeTestCase):
    def test_01_wrong_triggertype(self):
        with self.assertRaises(InvalidStreamTriggerType):
            StreamLoader(
                loader=Mock(),
                checkpoint_path="testpath",
                trigger_type="unknown",
            ).save(Mock())

    def test_02_processingtime(self):
        with self.assertRaises(NeedTriggerTimeWhenProcessingType):
            StreamLoader(
                loader=Mock(),
                checkpoint_path="testpath",
                trigger_type="processingtime",
            ).save(Mock())

    def test_03_wrong_output_mode(self):
        with self.assertRaises(UnknownStreamOutputMode):
            StreamLoader(
                loader=Mock(),
                outputmode="unknown",
                checkpoint_path="testpath",
                trigger_type="once",
            ).save(Mock())

    def test_04_checkpoint_overwrite_warning(self):
        """
        Tests that if an checkpoint path is provided
        as checkpoint_path,
        the user is warned, that it will overwrite the
        pre-existing checkpointLocation in the options dict,
        if any found.
        """

        with self.assertRaises(UserWarning):
            warnings.filterwarnings("error")

            StreamLoader(
                loader=Mock(),
                checkpoint_path="testpath",
                trigger_type="once",
                options_dict={"checkpointLocation": "anothertestpath"},
            ).save(Mock())



from datetime import timedelta

import pyspark.sql.functions as f
from pyspark.sql import DataFrame

from spetlr.eh import EventHubCapture
from spetlr.etl import Extractor
from spetlr.etl.extractors.simple_extractor import Readable


class IncrementalExtractor(Extractor):
    """This extractor will extract from any object that has a .read() method.
    Furthermore, it will use a target table for enabling incremental extraction.
    If needed it's possible to define a overlapping period using the timedelta function.
    NB: It is not recommended to use this on Eventhub data.
        Use EventHubCaptureExtractor instead.

    """

    def __init__(
        self,
        handle_source: Readable,
        handle_target: Readable,
        time_col_source: str,
        time_col_target: str,
        dataset_key: str = None,
        overlap_period: timedelta = None,
    ):
        super().__init__(dataset_key=dataset_key)
        self.handle_source = handle_source
        self.handle_target = handle_target
        self._timecol_source = time_col_source
        self._timecol_target = time_col_target
        self._overlap_period = overlap_period

    def read(self) -> DataFrame:
        if isinstance(self.handle_source, EventHubCapture):
            print(
                "It is recommended to use EventHubCaptureExtractor "
                "for extracting eventhub data."
                "EventHubCaptureExtractor is optimized for reading avro data."
            )

        df = self.handle_source.read()
        df_target = self.handle_target.read()

        # For incremental load, get the latest record from target table
        # In other words, get the maximum of the timestamp column
        target_max_time = (
            df_target.groupBy().agg(f.max(self._timecol_target)).collect()[0][0]
        )

        # If overlap_period is defined extract it from target_max_time
        if self._overlap_period and target_max_time is not None:
            target_max_time = target_max_time - self._overlap_period

        # If the target table is empty, target_max_time will be None
        # Only filter the input dataframe if the table is non-empty
        # table non-empty = target_max_time not None
        if target_max_time:
            df = df.where(f.col(self._timecol_source) > f.lit(target_max_time))

        return df



from spetlr.delta import DeltaHandle
from spetlr.eh.EventHubCaptureExtractor import EventHubCaptureExtractor
from spetlr.etl import EtlBase, Orchestrator
from spetlr.etl.loaders import SimpleLoader
from spetlr.orchestrators.eh2bronze.EhToDeltaBronzeTransformer import (
    EhToDeltaBronzeTransformer,
)
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaExtractor import (
    EhJsonToDeltaExtractor,
)


class EhToDeltaBronzeOrchestrator(Orchestrator):
    """

    This class has been designed to carry out the ETL task
    of ingest eventhub data to a bronze layer.

    Parameters:
    eh: EventHubCaptureExtractor for the eventhub data (raw)
    dh: DeltaHandle for the target delta table (bronze)

    Returns:
    Processed datasets of the super Orchestrator class
    """

    def __init__(self, eh: EventHubCaptureExtractor, dh: DeltaHandle):
        super().__init__()
        self.eh = eh
        self.dh = dh

        # step 1,
        #  - get the highest partition from the delta table,
        #  - truncate that partition
        #  - read the capture files from that partition
        self.extract_from(EhJsonToDeltaExtractor(eh, dh))

        # Step 2,
        # Transform the schema into a readable format
        self.transform_with(EhToDeltaBronzeTransformer(target_dh=dh))

        # the method filter_with can be used to insert any number of transformers here

        # final step,
        # append the rows to the delta table.
        self._loader = SimpleLoader(dh, mode="append")
        self.load_into(self._loader)

    @classmethod
    def from_tc(cls, eh_id: str, dh_id: str):
        return cls(
            eh=EventHubCaptureExtractor.from_tc(eh_id), dh=DeltaHandle.from_tc(dh_id)
        )

    def filter_with(self, etl: EtlBase):
        """Additional filters to execute before loading."""
        loader = self.steps.pop()

        # the following will fail if additional steps have been added.
        assert self._loader is loader, "unexpected change in etl steps"

        self.transform_with(etl)
        self.load_into(loader)



from typing import Protocol

from spetlr.tables import TableHandle


class ThMaker(Protocol):
    """Use this as a type when you need either DeltaHandle or a SqlServer instance."""

    def from_tc(self, id: str) -> TableHandle:
        pass



class SpetlrException(Exception):
    pass


class SpetlrKeyError(KeyError):
    pass


class NoTableException(SpetlrException):
    value = "No table found!"
    pass


class UnkownPathException(SpetlrException):
    value = "Something went wrong during reading of path!"
    pass


class ColumnDoesNotExistException(SpetlrException):
    pass


class MoreThanTwoDataFramesException(SpetlrException):
    pass


class EhJsonToDeltaException(SpetlrException):
    pass


class NoSuchSchemaException(SpetlrException):
    pass


class FalseSchemaDefinitionException(SpetlrException):
    pass


class UnregisteredSchemaDefinitionException(SpetlrException):
    pass


class NoRunId(SpetlrException):
    pass


class NoDbUtils(SpetlrException):
    pass


class NoSuchValueException(SpetlrKeyError):
    pass


class MissingUpsertJoinColumns(SpetlrKeyError):
    value = "You must specify upsert_join_cols"
    pass


class OnlyUseInSpetlrDebugMode(SpetlrKeyError):
    value = "Only call this if the configurator is in debug"


class InvalidStreamTriggerType(SpetlrException):
    value = "Trigger_type should either be {'availablenow', 'once', 'processingtime'}"
    pass


class NeedTriggerTimeWhenProcessingType(SpetlrException):
    value = "Trigger_time_seconds must be specified if trigger_type = 'processingtype'"
    pass


class UnknownStreamOutputMode(SpetlrException):
    value = (
        "Output mode should be one of the following ['complete', 'append', 'update']"
    )
    pass


class SparkVersionNotSupportedForSpetlrStreaming(SpetlrException):
    value = "Only DATABRICKS_RUNTIME_10_4 or above is supported for Spetlr streaming."
    pass



from pyspark.sql import DataFrame

from spetlr.configurator.configurator import Configurator
from spetlr.delta.delta_handle import DeltaHandleInvalidFormat
from spetlr.exceptions import SparkVersionNotSupportedForSpetlrStreaming
from spetlr.spark import Spark
from spetlr.tables import TableHandle


class FileHandle(TableHandle):
    def __init__(
        self,
        *,
        file_location: str,
        schema_location: str,
        data_format: str,
        options: dict = None
    ):
        """
        file_location: the path to the file location of the delta table
        schema_location: The location of the cloudfile schema.
            Databricks documentation uses the expected writeStream checkpoint_path
            as schema_location, see documentation at
            https://docs.databricks.com/getting-started/etl-quick-start.html#auto-loader

        data_format: The expected data format extracted by the Autoloader.
            Examples: JSON, CSV, PARQUET, AVRO, ORC, TEXT, and BINARYFILE.
            See: https://docs.databricks.com/ingestion/auto-loader/index.html
        """

        if Spark.version() < Spark.DATABRICKS_RUNTIME_10_4:
            raise SparkVersionNotSupportedForSpetlrStreaming()

        self._location = file_location
        self._data_format = data_format

        self._schema_location = schema_location
        self._options = options

        self._validate()

    @classmethod
    def from_tc(cls, id: str) -> "FileHandle":
        tc = Configurator()
        return cls(
            file_location=tc.table_property(id, "path", ""),
            data_format=tc.table_property(id, "format", ""),
            schema_location=tc.table_property(id, "schema_location", ""),
        )

    def _validate(self):
        """
        Validates the dataformat is not delta,
        since delta stream should use the DeltaHandle instead.
        """
        if self._data_format == "delta":
            raise DeltaHandleInvalidFormat("Use DeltaHandle.read_stream() for delta.")

    def read_stream(self) -> DataFrame:
        reader = (
            Spark.get()
            .readStream.format("cloudFiles")
            .option("cloudFiles.format", self._data_format)
            .option("cloudFiles.schemaLocation", self._schema_location)
        )

        if self._options is not None:
            reader = reader.options(**self._options)

        reader = reader.load(self._location)

        return reader



import importlib
import inspect
import sys
from typing import Any, Callable, Dict

ENTRY_POINT = "entry_point"


def main():
    """This function is entry point from which another entry point can be called.
    If you want to call this function:
    ```python
        def myfunction(myarg='default'): pass
    ```
    that is residing in the folder `mylib.myfolder.myfile`,
    then specify your task as follows:
    ```json
        "python_wheel_task": {
            "package_name": "spetlr",
            "entry_point": "spetlr_task",
            "named_parameters": {
                "entry_point": "mylib.myfolder.myfile:myfunction",
                "myarg": "myval"
            }
        }
    ```
    The named parameter 'entry_point' is mandatory.
    All arguments must be of type string.
    """

    kwargs = {}
    entry_point = None
    # sys.argv will contain strings like
    # [ "--entry_point=my.module:main", "--myarg=myval" ]
    for arg in sys.argv:
        # parameters of any other form are ignored
        if not arg.startswith("--"):
            continue
        if arg.find("=") < 0:
            continue

        k, v = arg[2:].split("=", 1)

        if k == ENTRY_POINT:
            # magic mandatory parameter
            entry_point = v
        else:
            # any other custom parameters
            kwargs[k] = v

    if entry_point is None:
        raise Exception("No entry_point specified.")

    # entry_point looks like "my.module:main"
    modname, qualname_separator, qualname = entry_point.partition(":")

    obj = importlib.import_module(modname)
    if qualname_separator:
        for attr in qualname.split("."):
            obj = getattr(obj, attr)

    kwargs = prepare_keyword_arguments(obj, kwargs)

    # call the callable with custom parameters
    return obj(**kwargs)


def prepare_keyword_arguments(callable_obj: Callable, kwargs_dict: Dict[str, Any]):
    """Reduce the dict down to a set of keys that the callable can actually be called
    with. Any extra keys are dropped with a warning."""
    signature = inspect.signature(callable_obj)
    parameters = signature.parameters

    if any(
        parameter.kind == inspect.Parameter.VAR_KEYWORD
        for parameter in parameters.values()
    ):
        # if any of the callable's parameters use the form **kwargs,
        # then we don't need to check further, just pass everything,
        return kwargs_dict

    valid_kwargs = {}
    for key, value in kwargs_dict.items():
        if key in parameters:  # only proceed if the parameter exists on the callable
            valid_kwargs[key] = value
        else:
            print(
                f"WARNING: Ignoring job parameter: {key}. "
                "The entry point cannot receive it."
            )

    return valid_kwargs



"""
Pyspark local sessions on windows insist on finding the python interpreter
as python3 on the PATH.
Regardless of settings such as spark.pyspark.python which have no apparent effect.
This fake script is added to the path as "python3" in you virtual environment
and calls the same interpreter as the "python" executable.
This makes spark testing work on windows using local sessions.
"""
import os
import sys


def python3():
    os.execv(sys.executable, sys.argv)



import unittest

from spetlrtools.testing import TestHandle

from spetlr.etl import Orchestrator
from spetlr.etl.extractors import SimpleExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.spark import Spark
from spetlr.transformers import UnionTransformer


class MergeDfIntoTargetTest(unittest.TestCase):
    def test_01_union(self):
        schema = "i integer, s string"

        th1 = TestHandle(provides=Spark.get().createDataFrame([(1, "a")], schema))
        th2 = TestHandle(provides=Spark.get().createDataFrame([(1, "a")], schema))

        th3 = TestHandle()

        o = Orchestrator()
        o.extract_from(SimpleExtractor(th1, "th1"))
        o.extract_from(SimpleExtractor(th2, "th2"))
        o.transform_with(UnionTransformer())
        o.load_into(SimpleLoader(th3))
        o.execute()

        self.assertEqual(th1.provides.schema, th3.overwritten.schema)



import uuid
import warnings

from pyspark.sql import DataFrame
from pyspark.sql.streaming import DataStreamWriter

from spetlr.etl import Loader
from spetlr.exceptions import (
    InvalidStreamTriggerType,
    NeedTriggerTimeWhenProcessingType,
    SparkVersionNotSupportedForSpetlrStreaming,
    UnknownStreamOutputMode,
)
from spetlr.spark import Spark


class StreamLoader(Loader):
    def __init__(
        self,
        *,
        loader: Loader,
        checkpoint_path: str = None,
        options_dict: dict = None,
        trigger_type: str = "availablenow",
        trigger_time_seconds: int = None,
        outputmode: str = "update",
        query_name: str = None,
        await_termination: bool = False,
    ):
        """
        loader: A SPETLR Loader
        checkpoint_path: The location of the checkpoints, <table_name>/_checkpoints
            The Delta Lake VACUUM function removes all files not managed by Delta Lake
            but skips any directories that begin with _. You can safely store
            checkpoints alongside other data and metadata for a Delta table
            using a directory structure such as <table_name>/_checkpoints
            See: https://docs.databricks.com/structured-streaming/delta-lake.html
        trigger_type: The trigger type of the stream.
            See: https://docs.databricks.com/structured-streaming/triggers.html
        trigger_time: If the trigger has is "processingtime",
            it should have a trigger time associated
        outputmode: The streaming output mode: append, complete or update
            See: https://spark.apache.org/docs/3.1.1/api/python/reference/api
                /pyspark.sql.streaming.DataStreamWriter.outputMode.html
        query_name: The name of the stream query (defaults to an uuid)
        await_termination: If true, then waits for the termination of THIS query
            See: https://spark.apache.org/docs/3.1.1/api/python
                 /reference/api/pyspark.sql.streaming.StreamingQuery.awaitTermination.html


        """

        super().__init__()
        self._loader = loader
        self._options_dict = options_dict
        self._outputmode = outputmode
        self._trigger_type = trigger_type
        self._trigger_time_seconds = trigger_time_seconds
        self._query_name = query_name or str(
            uuid.uuid4().hex
        )  # Consider if this is smart?
        self._checkpoint_path = checkpoint_path
        self._await_termination = await_termination
        self._validate_checkpoint()

        if Spark.version() < Spark.DATABRICKS_RUNTIME_10_4:
            raise SparkVersionNotSupportedForSpetlrStreaming()

        # Set checkpoint path always
        self._options_dict = self._options_dict or {}

        if "checkpointLocation" in self._options_dict:
            warnings.warn(
                "There was found a pre-existing checkpointLocation in the"
                "options dict. Overwrites the location with the \n"
                "provided location from checkpoint_path..."
            )

        self._options_dict["checkpointLocation"] = self._checkpoint_path

        # "continuous" is not available when using foreachBatch()
        # https://docs.databricks.com/structured-streaming/foreach.html#apply-additional-dataframe-operations

        if self._trigger_type not in {"availablenow", "once", "processingtime"}:
            raise InvalidStreamTriggerType()

        if (self._trigger_type == "processingtime") and (
            self._trigger_time_seconds is None
        ):
            raise NeedTriggerTimeWhenProcessingType()

        if self._outputmode not in {"complete", "append", "update"}:
            raise UnknownStreamOutputMode()

    def save(self, df: DataFrame) -> None:
        df_stream = (
            df.writeStream.options(**self._options_dict)
            .outputMode(self._outputmode)
            .queryName(self._query_name)
        )

        df_stream = self._add_trigger_type(df_stream)

        df_stream.foreachBatch(self._foreachbatch)

        query = df_stream.start()

        if self._await_termination:
            query.awaitTermination()

    def _foreachbatch(self, df: DataFrame, _: int = None):
        self._loader.save(df)

    def _add_trigger_type(self, writer: DataStreamWriter):
        if self._trigger_type == "availablenow":
            return writer.trigger(availableNow=True)
        elif self._trigger_type == "once":
            return writer.trigger(once=True)
        elif self._trigger_type == "processingtime":
            return writer.trigger(
                processingTime=f"{self._trigger_time_seconds} seconds",
            )
        else:
            raise ValueError("Unknown trigger type.")

    def _validate_checkpoint(self):
        if "/_" not in self._checkpoint_path:
            print(
                "RECOMMENDATION: You can safely store checkpoints alongside "
                "other data and metadata for a Delta table using a directory "
                "structure such as <table_name>/_checkpoints"
            )



import unittest

from pyspark.sql.types import IntegerType, StructField, StructType
from spetlrtools.testing import DataframeTestCase

from spetlr.etl.log.log_transformers import NullLogTransformer
from spetlr.utils import DataframeCreator


class TestNullLogTransformer(DataframeTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        schema = StructType(
            [
                StructField("col_1", IntegerType(), True),
                StructField("col_2", IntegerType(), True),
                StructField("col_3", IntegerType(), True),
            ]
        )

        data_1 = [
            (1, None, 1),
            (2, None, 2),
            (3, 3, 3),
            (4, 4, 4),
            (5, 5, 5),
        ]

        data_2 = [
            (1, 1, 1),
            (2, 2, 2),
            (3, 3, 3),
        ]

        df_1 = DataframeCreator.make(schema, data_1)
        df_2 = DataframeCreator.make(schema, data_2)

        cls.dataset = {"df_1": df_1, "df_2": df_2}

    def test_log_01(self) -> None:
        df = NullLogTransformer(
            log_name="test_log_name",
            column_name="col_2",
        ).log(self.dataset["df_1"])

        self.assertDataframeMatches(
            df=df,
            columns=["Count", "NumberOfNulls", "PercentageOfNulls", "ColumnName"],
            expected_data=[(5, 2, "0.4000", "col_2")],
        )

    def test_log_many_02(self) -> None:
        df = NullLogTransformer(
            log_name="test_log_name",
            column_name="col_2",
        ).log_many(self.dataset)

        self.assertDataframeMatches(
            df=df,
            columns=["Count", "NumberOfNulls", "PercentageOfNulls", "ColumnName"],
            expected_data=[
                (5, 2, "0.4000", "col_2"),
                (3, 0, "0.0000", "col_2"),
            ],
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from spetlrtools.testing import DataframeTestCase

from spetlr import Configurator
from spetlr.deltaspec import DeltaTableSpec
from spetlr.spark import Spark
from tests.cluster.delta.deltaspec import tables


@unittest.skipUnless(
    Spark.version() >= Spark.DATABRICKS_RUNTIME_11_3,
    "Drop column only supported from DBR 11.0",
)
class TestTableSpec(DataframeTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        c = Configurator()
        c.clear_all_configurations()
        c.set_debug()

        c.register("mydb", dict(name="myDeltaTableSpecTestDb{ID}"))
        cls.base = tables.base
        cls.target = tables.target

        db = c.get("mydb", "name")
        Spark.get().sql(f"CREATE DATABASE {db};")

    @classmethod
    def tearDownClass(cls) -> None:
        c = Configurator()
        db = c.get("mydb", "name")
        # clean up after test.
        Spark.get().sql(f"DROP DATABASE {db} CASCADE")

    def test_01_tblspec(self):
        # at first the table does not exist
        diff = self.base.compare_to_name()
        self.assertTrue(diff.is_different(), diff)
        self.assertTrue(diff.nullbase(), diff)

        # then we make it exist
        self.base.make_storage_match()

        # now it exists and matches
        diff = self.base.compare_to_name()
        self.assertFalse(diff.is_different(), repr(diff))

        # but it does not match the target
        diff = self.target.compare_to_name()
        # the names are the same
        self.assertTrue(diff.name_match(), repr(diff))
        # the rest of the table is not the same
        self.assertTrue(diff.is_different(), repr(diff))

        # the table is not readable because of schema mismatch
        self.assertFalse(self.target.is_readable())

        # overwriting is possible and updates to the target schema
        df = Spark.get().createDataFrame([(1, "a", 3.14, "b", "c")], self.target.schema)
        self.target.make_storage_match(errors_as_warnings=True)

        self.target.get_dh().overwrite(df, overwriteSchema=True)

        # now the base no longer matches
        diff = self.base.compare_to_name()
        self.assertTrue(diff.is_different(), diff)

        # but the target matches.
        diff = self.target.compare_to_name()
        self.assertFalse(diff.is_different(), repr(diff))

    def test_02_location_change(self):
        tables.oldlocation.make_storage_match(allow_table_create=True)

        # location mismatch makes the tables not readable
        self.assertFalse(tables.newlocation.is_readable())

        # overwriting will update the table location.
        tables.newlocation.make_storage_match(allow_location_change=True)

        diff = tables.newlocation.compare_to_name()
        self.assertTrue(diff.complete_match(), diff)
        self.assertTrue(tables.newlocation.is_readable(), diff)

    def test_03_name_change(self):
        tables.oldname.make_storage_match(allow_table_create=True)

        spark = Spark.get()
        df = spark.createDataFrame([("eggs", 3.5, "spam")], tables.oldname.schema)
        tables.oldname.get_dh().overwrite(df)

        for stmt in tables.newname.compare_to(tables.oldname).alter_statements(
            allow_name_change=True
        ):
            spark.sql(stmt)

        diff = tables.newname.compare_to_name()
        self.assertTrue(diff.complete_match(), diff)
        self.assertEqual(tables.newname.get_dh().read().count(), 1)

    def test_04_can_control_managed_table(self):
        tables.managed.make_storage_match(allow_table_create=True)

        diff = tables.managed.compare_to_name()
        self.assertTrue(diff.complete_match(), diff)

    def test_05_modify_to_allow_drop(self):
        # a table has been created without the DeltaTableSpec:
        spark = Spark.get()
        spark.sql(tables.simple_create_sql.format(**Configurator().get_all_details()))

        # now a modified version is to be applied that drops columns:
        ds = DeltaTableSpec.from_sql(tables.simple_modified_sql)
        ds.make_storage_match(allow_columns_drop=True)

        diff = ds.compare_to_name()
        self.assertTrue(diff.complete_match(), diff)



from spetlr.exceptions import SpetlrException


class InvalidSpecificationError(SpetlrException):
    """Internally this exception is used when attempting
    to unpack configurations into table and db specifications and this fails"""


class NoTableAtTarget(InvalidSpecificationError):
    """Indiactes thate there is not table at the target location."""


class TableSpectNotEnforcable(SpetlrException):
    """For some reason it was not possible to make the storage match the spec."""


class TableSpecNotReadable(SpetlrException):
    """A read was attempted, but the table is not readable as specified."""


class TableSpecSchemaMismatch(SpetlrException):
    """A write was attempted, but the dataframe is not as specified."""


class InvalidTableName(SpetlrException):
    """A table name of the given components is invalid."""



from spetlrtools.testing import DataframeTestCase

from spetlr.filehandle import FileHandle


class FileHandleTests(DataframeTestCase):
    def test_01_can_set_options(self):
        """
        Simple test that options attribute is set in the FileHandle
        """

        class TestClass(FileHandle):
            def __init__(
                self,
                *,
                file_location: str,
                schema_location: str,
                data_format: str,
                options: dict
            ):
                super().__init__(
                    file_location=file_location,
                    schema_location=schema_location,
                    data_format=data_format,
                    options=options,
                )

            def print(self) -> int:
                return 1

        tester = TestClass(
            file_location="test",
            schema_location="test",
            data_format="csv",
            options={"somesetting": "somevalue"},
        )

        self.assertEqual(tester._options, {"somesetting": "somevalue"})



import uuid
from typing import List

import pyspark.sql.functions as F
import pyspark.sql.types as T
from pyspark.sql import DataFrame

from spetlr.etl import Transformer


class GenerateMd5ColumnTransformer(Transformer):
    """
    This transformer generates a unique column with md5 encoding based on other columns.
    The transformer also handles if a value is NULL, by replacing it with empty string.

    Attributes:
    ----------
        col_name : String
            name of the column to put the md5 encoding
        col_list : List[str]
            list of columns to use for the md5 encoding
        dataset_input_keys : Union[str, List[str]]
            list of input dataset keys
        dataset_output_key : str
            output dataset key
    """

    def __init__(
        self,
        *,
        col_name: str,
        col_list: List[str],
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.col_name = col_name
        self.col_list = col_list

    def process(self, df: DataFrame) -> DataFrame:
        temp_col_list = []
        for col in self.col_list:
            # Place null safe value in temp column
            temp_col_name = str(uuid.uuid4())
            temp_col_list.append(temp_col_name)

            # Cast column to string
            # Coalesce with empty string in order to ensure md5 enablement
            _query = F.coalesce(F.col(col).cast(T.StringType()), F.lit(""))

            df = df.withColumn(temp_col_name, _query)

        df = df.withColumn(self.col_name, F.md5(F.concat(*temp_col_list)))

        df = df.drop(*temp_col_list)

        return df



from dataclasses import dataclass
from typing import Dict, List

ENUM_ASC = 1
ENUM_DESC = 2


@dataclass
class SortingCol:
    name: str
    ordering: int = ENUM_ASC


@dataclass
class ClusteredBy:
    clustering_cols: List[str]
    n_buckets: int = 0
    sorting: List[SortingCol] = None


@dataclass
class StatementBlocks:
    schema: str = None
    using: str = "DELTA"
    options: Dict[str, str] = None
    partitioned_by: List[str] = None
    clustered_by: ClusteredBy = None
    location: str = None
    comment: str = None
    tblproperties: Dict[str, str] = None
    dbproperties: Dict[str, str] = None

    def get_simple_structure(self):
        object_details = {
            "path": self.location,
            "format": self.using.lower() if self.using else None,
            "options": self.options,
            "partitioned_by": self.partitioned_by,
            "comment": self.comment,
            "tblproperties": self.tblproperties,
            "dbproperties": self.dbproperties,
        }

        if self.schema:
            self.schema = self.schema.strip()
            if self.schema[0] == "(":
                self.schema = self.schema[1:-1]
            object_details["schema"] = {"sql": self.schema}
            object_details["_raw_sql_schema"] = self.schema

        if self.clustered_by:
            object_details["clustered_by"] = {
                "cols": self.clustered_by.clustering_cols,
                "buckets": self.clustered_by.n_buckets,
            }
            if self.clustered_by.sorting:
                object_details["clustered_by"]["sorted_by"] = []
                for col in self.clustered_by.sorting:
                    object_details["clustered_by"]["sorted_by"].append(
                        dict(
                            name=col.name,
                            ordering=("ASC" if col.ordering == ENUM_ASC else "DESC"),
                        )
                    )

        return {k: v for k, v in object_details.items() if v is not None}



import time
import unittest

from pyspark.sql.utils import AnalysisException

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl import Orchestrator
from spetlr.etl.extractors.stream_extractor import StreamExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.loaders.stream_loader import StreamLoader
from spetlr.spark import Spark
from spetlr.testutils.stop_test_streams import stop_test_streams


@unittest.skipUnless(
    Spark.version() >= Spark.DATABRICKS_RUNTIME_10_4,
    f"Spetlr Streaming not available for Spark version {Spark.version()}",
)
class DeltaStreamTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        Configurator().clear_all_configurations()
        Configurator().set_debug()

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("MyDb").drop_cascade()

        stop_test_streams()

    def test_01_configure(self):
        tc = Configurator()
        tc.register(
            "MyDb", {"name": "TestDb{ID}", "path": "/mnt/spetlr/silver/testdb{ID}"}
        )

        tc.register(
            "MyTbl",
            {
                "name": "TestDb{ID}.TestTbl",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/_checkpoint_path_tbl",
                "query_name": "testquerytbl{ID}",
            },
        )

        mirror_cp_path = "/mnt/spetlr/silver/testdb{ID}/_checkpoint_path_tblmirror"
        tc.register(
            "MyTblMirror",
            {
                "name": "TestDb{ID}.TestTblMirror",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtblmirror",
                "format": "delta",
                "checkpoint_path": mirror_cp_path,
                "await_termination": True,
                "query_name": "testquerytblmirror{ID}",
            },
        )

        tc.register(
            "MyTbl2",
            {
                "name": "TestDb{ID}.TestTbl2",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/"
                "_checkpoint_path_tbl2",
                "query_name": "testquerytbl2{ID}",
            },
        )

        tc.register(
            "MyTbl3",
            {
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl3",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/"
                "_checkpoint_path_tbl3",
                "await_termination": True,
                "query_name": "testquerytbl3{ID}",
            },
        )

        tc.register(
            "MyTbl4",
            {
                "name": "TestDb{ID}.TestTbl4",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl4",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/"
                "_checkpoint_path_tbl4",
                "query_name": "testquerytbl4{ID}",
            },
        )

        tc.register(
            "MyTbl5",
            {
                "name": "TestDb{ID}.TestTbl5",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl5",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}"
                "/_checkpoint_path_tbl5",
                "query_name": "testquerytbl5{ID}",
            },
        )

        # test instantiation without error
        DbHandle.from_tc("MyDb")
        DeltaHandle.from_tc("MyTbl")
        DeltaHandle.from_tc("MyTblMirror")
        DeltaHandle.from_tc("MyTbl2")
        DeltaHandle.from_tc("MyTbl3")
        DeltaHandle.from_tc("MyTbl4")
        DeltaHandle.from_tc("MyTbl5")

    def test_02_write_data_with_deltahandle(self):
        self._overwrite_two_rows_to_table("MyTbl")

    def test_03_create(self):
        db = DbHandle.from_tc("MyDb")
        db.create()

        dh = DeltaHandle.from_tc("MyTbl")
        dh.create_hive_table()

        # test hive access:
        df = dh.read()
        self.assertEqual(2, df.count())

    def test_04_read(self):
        df = DeltaHandle.from_tc("MyTbl").read_stream()
        self.assertTrue(df.isStreaming)

    def test_05_truncate(self):
        dsh = DeltaHandle.from_tc("MyTbl")
        dsh.truncate()

        result = DeltaHandle.from_tc("MyTbl").read()
        self.assertEqual(0, result.count())

    def test_06_etl(self):
        self._overwrite_two_rows_to_table("MyTbl")
        self._create_tbl_mirror()

        dh = DeltaHandle.from_tc("MyTbl")
        dh_target = DeltaHandle.from_tc("MyTblMirror")

        o = Orchestrator()
        o.extract_from(StreamExtractor(dh, dataset_key="MyTbl"))
        o.load_into(
            StreamLoader(
                loader=SimpleLoader(dh_target, mode="append"),
                await_termination=True,
                checkpoint_path=Configurator().get("MyTblMirror", "checkpoint_path"),
                query_name=Configurator().get("MyTblMirror", "query_name"),
            )
        )
        o.execute()

        result = DeltaHandle.from_tc("MyTblMirror").read()
        self.assertEqual(2, result.count())

    def test_07_write_path_only(self):
        self._overwrite_two_rows_to_table("MyTbl")
        # check that we can write to the table with no "name" property
        dh1 = DeltaHandle.from_tc("MyTbl")

        dh3 = DeltaHandle.from_tc("MyTbl3")

        # dsh3.append(ah, mergeSchema=True)

        o = Orchestrator()
        o.extract_from(StreamExtractor(dh1, dataset_key="MyTbl"))
        o.load_into(
            StreamLoader(
                loader=SimpleLoader(dh3, mode="append"),
                await_termination=True,
                checkpoint_path=Configurator().get("MyTbl3", "checkpoint_path"),
                query_name=Configurator().get("MyTbl3", "query_name"),
            ),
        )
        o.execute()

        # Read data from mytbl3
        result = dh3.read()
        self.assertEqual(2, result.count())

    def test_09_trigger_once(self):
        self._overwrite_two_rows_to_table("MyTbl")
        # check that we can write to the table with no "name" property
        dh1 = DeltaHandle.from_tc("MyTbl")

        dh3 = DeltaHandle.from_tc("MyTbl3")

        # dsh3.append(ah, mergeSchema=True)

        o = Orchestrator()
        o.extract_from(StreamExtractor(dh1, dataset_key="MyTbl"))
        o.load_into(
            StreamLoader(
                loader=SimpleLoader(dh3, mode="append"),
                await_termination=False,
                checkpoint_path=Configurator().get("MyTbl3", "checkpoint_path"),
                trigger_type="once",
                query_name=Configurator().get("MyTbl3", "query_name"),
            ),
        )
        o.execute()

        # wait 20 sec for the stream to start
        time.sleep(20)

        stop_test_streams()

    def test_09_trigger_processing_time(self):
        self._overwrite_two_rows_to_table("MyTbl")
        # check that we can write to the table with no "name" property
        dh1 = DeltaHandle.from_tc("MyTbl")

        dh3 = DeltaHandle.from_tc("MyTbl3")

        # dsh3.append(ah, mergeSchema=True)

        o = Orchestrator()
        o.extract_from(StreamExtractor(dh1, dataset_key="MyTbl"))
        o.load_into(
            StreamLoader(
                loader=SimpleLoader(dh3, mode="append"),
                options_dict={},
                await_termination=False,
                checkpoint_path=Configurator().get("MyTbl3", "checkpoint_path"),
                trigger_type="processingtime",
                trigger_time_seconds=5,
                query_name=Configurator().get("MyTbl3", "query_name"),
            ),
        )
        o.execute()

        # wait 60 sec for the stream to start
        time.sleep(60)

        stop_test_streams()

    def test_10_delete(self):
        dh = DeltaHandle.from_tc("MyTbl")
        dh.drop_and_delete()
        with self.assertRaises(AnalysisException):
            dh.read()

    def _overwrite_two_rows_to_table(self, tblid: str):
        dh = DeltaHandle.from_tc(tblid)

        df = Spark.get().createDataFrame([(1, "a"), (2, "b")], "id int, name string")

        dh.overwrite(df, mergeSchema=True)

    def _create_tbl_mirror(self):
        dh = DeltaHandle.from_tc("MyTblMirror")
        Spark.get().sql(
            f"""
                            CREATE TABLE {dh.get_tablename()}
                            (
                            id int,
                            name string
                            )
                            LOCATION '{Configurator().get("MyTblMirror","path")}'
                        """
        )



from typing import List

import pycountry
import pyspark.sql.functions as F
import pyspark.sql.types as T
from pyspark.sql import DataFrame

from spetlr.etl import Transformer


def translate_country_to_alpha2(country_name: str) -> str:
    """
    Method to translate country names to alpha-2 codes
    Args:
        country_name: The name of the country
    Returns:
        The corresponding alpha-2 code or an empty string if an invalid name was given
    """
    country = pycountry.countries.get(name=country_name)
    if country:
        return country.alpha_2
    else:
        return ""


translateUDF = F.udf(lambda z: translate_country_to_alpha2(z), T.StringType())


class CountryToAlphaCodeTransformer(Transformer):
    def __init__(
        self,
        col_name: str,
        output_col_name: str = None,
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ) -> None:
        """
        A simple transformer to translate country names to alpha-2 codes

        Args:
            col_name: The name of the column with country names
            output_col_name: The name of the column to create,
            defaults to col_name if none is given
        """
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.col_name = col_name
        self.output_col_name = output_col_name or self.col_name

    def process(self, df: DataFrame) -> DataFrame:
        """
        Method to add column to passed dataframe 'df'
        Args:
            df: The dataframe to work on
        Returns:
            The dataframe with an added or overwritten column
        """
        df = df.withColumn(self.output_col_name, translateUDF(F.col(self.col_name)))
        return df



from .incremental_extractor import IncrementalExtractor
from .lazy_extractor import LazyExtractor
from .lazy_simple_extractor import LazySimpleExtractor
from .simple_extractor import SimpleExtractor
from .stream_extractor import StreamExtractor

__all__ = [
    "IncrementalExtractor",
    "SimpleExtractor",
    "StreamExtractor",
    "LazyExtractor",
    "LazySimpleExtractor",
]



import json
from unittest.mock import Mock, patch

from pyspark.sql.types import (
    BinaryType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.testing import DataframeTestCase
from spetlrtools.testing.TestHandle import TestHandle
from spetlrtools.time import dt_utc

from spetlr.orchestrators import (
    EhToDeltaBronzeOrchestrator,
    EhToDeltaSilverOrchestrator,
)
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaExtractor import (
    EhJsonToDeltaExtractor,
)
from spetlr.spark import Spark
from spetlr.utils import DataframeCreator


class EhtoBronzeAndSilverUnitTests(DataframeTestCase):
    """
    This tests both the bronze and silver eventhub orchestrator.

    """

    _target_schema_bronze = StructType(
        [
            StructField("EventhubRowId", LongType(), True),
            StructField("BodyId", LongType(), True),
            StructField("Body", StringType(), True),
            StructField("EnqueuedTimestamp", TimestampType(), True),
            StructField("StreamingTime", TimestampType(), True),
            StructField("SequenceNumber", LongType(), True),
            StructField("Offset", StringType(), True),
            StructField("SystemProperties", StringType(), True),
            StructField("Properties", StringType(), True),
            StructField("pdate", TimestampType(), True),
        ]
    )

    _target_cols_bronze = [
        "EventhubRowId",
        "BodyId",
        "Body",
        "EnqueuedTimestamp",
        "StreamingTime",
        "SequenceNumber",
        "Offset",
        "SystemProperties",
        "Properties",
        "pdate",
    ]

    _target_schema_silver = StructType(
        [
            StructField("Id", StringType(), True),
            StructField("EnqueuedTimestamp", TimestampType(), True),
            StructField("pdate", TimestampType(), True),
        ]
    )

    _target_cols_silver = ["Id", "EnqueuedTimestamp", "pdate"]

    _capture_eventhub_output_schema = StructType(
        [
            StructField("SequenceNumber", LongType(), True),
            StructField("Offset", StringType(), True),
            StructField("SystemProperties", StringType(), True),
            StructField("Properties", StringType(), True),
            StructField("Body", BinaryType(), True),
            StructField("pdate", TimestampType(), True),
            StructField("EnqueuedTimestamp", TimestampType(), True),
        ]
    )

    def test_01_test_bronze_and_silver(self):
        eh_mock = Mock()

        dh_bronze = TestHandle(
            provides=DataframeCreator.make_partial(
                self._target_schema_bronze, self._target_cols_bronze, []
            )
        )

        dh_silver = TestHandle(
            provides=DataframeCreator.make_partial(
                self._target_schema_silver, self._target_cols_silver, []
            )
        )

        df_in = Spark.get().createDataFrame(
            [
                (
                    22222,
                    "OffsetTest",
                    "SystemPropertiesTest",
                    "PropertiesTest",
                    json.dumps(
                        {
                            "Id": "1234",
                            "name": "John",
                        }
                    ).encode(
                        "utf-8"
                    ),  # Body
                    dt_utc(2021, 10, 31, 0, 0, 0),  # pdate
                    dt_utc(2021, 10, 31, 0, 0, 0),  # EnqueuedTimestamp
                ),
            ],
            self._capture_eventhub_output_schema,
        )

        expected_silver = [
            (
                "1234",  # Id
                dt_utc(2021, 10, 31, 0, 0, 0),  # EnqueuedTimestamp
                dt_utc(2021, 10, 31, 0, 0, 0),  # pdate
            ),
        ]

        with patch.object(EhJsonToDeltaExtractor, "read", return_value=df_in) as p1:
            bronze_orchestrator = EhToDeltaBronzeOrchestrator(eh=eh_mock, dh=dh_bronze)

            bronze_orchestrator.execute()

            dh_bronze_mock = Mock()
            dh_bronze_mock.read = Mock(return_value=dh_bronze.appended)

            silver_orchestrator = EhToDeltaSilverOrchestrator(
                dh_source=dh_bronze_mock, dh_target=dh_silver, upsert_join_cols=["id"]
            )
            silver_orchestrator.execute()

            df_result = dh_silver.upserted

            self.assertDataframeMatches(
                df_result,
                ["Id", "EnqueuedTimestamp", "pdate"],
                expected_silver,
            )
            p1.assert_called_once()



import time
import unittest
from datetime import datetime, timedelta, timezone

from pyspark.sql import DataFrame
from spetlrtools.time import dt_utc

from spetlr import Configurator
from spetlr.delta import DeltaHandle
from spetlr.eh import EventHubJsonPublisher
from spetlr.eh.EventHubCaptureExtractor import EventHubCaptureExtractor
from spetlr.etl import Transformer

# from spetlr.functions import init_dbutils
from spetlr.orchestrators import EhJsonToDeltaOrchestrator
from spetlr.spark import Spark
from tests.cluster.values import resourceName

from .SpetlrEh import SpetlrEh


class EventHubsTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        Configurator().clear_all_configurations()

    def test_01_publish_and_read(self):
        eh = SpetlrEh()

        df = Spark.get().createDataFrame([(1, "a"), (2, "b")], "id int, name string")
        publisher = EventHubJsonPublisher(eh)
        publisher.save(df)

        time.sleep(100)  # just wait the EH captures once a minute anyway.

        tc = Configurator()
        tc.register(
            "SpetlrEh",
            {
                "path": f"/mnt/{resourceName()}/silver/{resourceName()}/spetlreh",
                "format": "avro",
                "partitioning": "ymd",
            },
        )
        eh = EventHubCaptureExtractor.from_tc("SpetlrEh")
        df = eh.read()
        self.assertTrue(df.count(), 2)

        df = eh.read(
            (datetime.now() - timedelta(hours=24)).replace(
                hour=0, minute=0, second=0, microsecond=0
            )
        )
        self.assertTrue(df.count(), 2)

        df = eh.read().select("EnqueuedTimestamp")
        self.assertEqual(
            df.schema.fields[0].dataType.typeName().upper(),
            "TIMESTAMP",
        )
        row_written: datetime = df.take(1)[0][0]
        # assert that the eventhub row was written less than 1000 seconds ago
        self.assertLess(
            abs((row_written.astimezone(timezone.utc) - dt_utc()).total_seconds()), 1000
        )

        # def test_05_eh_json_orchestrator(self):
        # the orchestrator has a complex functionality that can only be fully tested
        # on a substantial holding of capture files. That is not possible here, but
        # such tests were carried out during development.
        # The situation here only tests the basic functions.

        # Part 1, YMD partitioned
        tc = Configurator()
        tc.set_debug()
        tc.register("CpTblYMD", {"name": "CaptureTableYMD{ID}"})
        Spark.get().sql(
            f"""
            CREATE TABLE {tc.table_name('CpTblYMD')}
            (
                id int,
                name string,
                y int,
                m int,
                d int
            )
            PARTITIONED BY (y,m,d)
        """
        )

        eh_orch = EhJsonToDeltaOrchestrator.from_tc("SpetlrEh", "CpTblYMD")
        eh_orch.execute()

        df = DeltaHandle.from_tc("CpTblYMD").read().select("id", "name")

        rows = {tuple(row) for row in df.collect()}
        self.assertEqual({(1, "a"), (2, "b")}, rows)

        # Part 2, pdate partitioned.

        tc.register("CpTblDate", {"name": "CaptureTableDate{ID}"})
        Spark.get().sql(
            f"""
            CREATE TABLE {tc.table_name('CpTblDate')}
            (
                id INTEGER,
                name STRING,
                pdate TIMESTAMP
            )
            PARTITIONED BY (pdate)
        """
        )

        # test the insertion of additional filters
        class IdFilter(Transformer):
            def process(self, df: DataFrame) -> DataFrame:
                return df.filter("id>1")

        eh_orch2 = EhJsonToDeltaOrchestrator.from_tc("SpetlrEh", "CpTblDate")
        eh_orch2.filter_with(IdFilter())
        eh_orch2.execute()

        df2 = DeltaHandle.from_tc("CpTblDate").read().select("id", "name")

        rows = {tuple(row) for row in df2.collect()}
        self.assertEqual({(2, "b")}, rows)



from datetime import date, datetime
from typing import Any, Dict, List, Optional, Union

from pyspark.sql import DataFrame

from spetlr.configurator.configurator import Configurator
from spetlr.exceptions import SpetlrException
from spetlr.functions import get_unique_tempview_name, init_dbutils
from spetlr.spark import Spark
from spetlr.tables.TableHandle import TableHandle
from spetlr.utils.CheckDfMerge import CheckDfMerge
from spetlr.utils.GetMergeStatement import GetMergeStatement


class DeltaHandleException(SpetlrException):
    pass


class DeltaHandleInvalidName(DeltaHandleException):
    pass


class DeltaHandleInvalidFormat(DeltaHandleException):
    pass


class DeltaHandle(TableHandle):
    def __init__(
        self,
        name: str,
        location: str = None,
        data_format: str = "delta",
        options_dict: Dict[str, str] = None,
        ignore_changes: bool = True,
        stream_start: Union[datetime, str] = None,
        max_bytes_per_trigger: int = None,
    ):
        """
        name: The name of the Delta table.
        location (optional): The file-system path to the Delta table files.
        data_format (optional): Always delta-format. Todo: Remove in future PR.
        options_dict (optional): All other string options for pyspark.
        ignore_changes (optional): ignore transactions that delete data
                                    at partition boundaries.
        stream_start (optional):  If string format, it accepts anything that
                                  the `dateparser` library can parse.
        max_bytes_per_trigger (optional): How much data gets
                                processed in each micro-batch.
        """
        self._name = name
        self._location = location
        self._data_format = data_format

        self._partitioning: Optional[List[str]] = None
        self._validate()

        if options_dict is None or options_dict == "":
            self.set_options_dict({})
        else:
            self.set_options_dict(options_dict)

        self._options_dict.update({"ignoreChanges": str(ignore_changes)})

        if stream_start and stream_start != "":
            if isinstance(stream_start, datetime):
                self._options_dict["startingTimestamp"] = stream_start.strftime(
                    "%Y-%m-%dT%H:%M:%S.%fZ"
                )
            else:
                self._options_dict["startingTimestamp"] = stream_start

        if max_bytes_per_trigger and max_bytes_per_trigger != "":
            self._options_dict["maxBytesPerTrigger"] = str(max_bytes_per_trigger)

    @classmethod
    def from_tc(cls, id: str) -> "DeltaHandle":
        tc = Configurator()
        return cls(
            name=tc.table_property(id, "name", ""),
            location=tc.table_property(id, "path", ""),
            data_format=tc.table_property(id, "format", "delta"),
            ignore_changes=tc.table_property(id, "ignore_changes", "True"),
            stream_start=tc.table_property(id, "stream_start", ""),
            max_bytes_per_trigger=tc.table_property(id, "max_bytes_per_trigger", ""),
        )

    def _validate(self):
        """Validates that the name is either db.table or just table."""
        if not self._name:
            if not self._location:
                raise DeltaHandleInvalidName(
                    "Cannot create DeltaHandle without name or path"
                )
            self._name = f"delta.`{self._location}`"
        else:
            name_parts = self._name.split(".")
            if len(name_parts) == 1:
                self._db = None
                self._table_name = name_parts[0]
            elif len(name_parts) == 2:
                self._db = name_parts[0]
                self._table_name = name_parts[1]
            elif len(name_parts) == 3:
                self._catalog = name_parts[0]
                self._db = name_parts[1]
                self._table_name = name_parts[2]
            else:
                raise DeltaHandleInvalidName(f"Could not parse name {self._name}")

        # only format delta is supported.
        if self._data_format != "delta":
            raise DeltaHandleInvalidFormat("Only format delta is supported.")

    def read(self) -> DataFrame:
        """Read table by path if location is given, otherwise from name."""
        if self._location:
            return Spark.get().read.format(self._data_format).load(self._location)
        return Spark.get().table(self._name)

    def write_or_append(
        self,
        df: DataFrame,
        mode: str,
        mergeSchema: bool = None,
        overwriteSchema: bool = None,
    ) -> None:
        assert mode in {"append", "overwrite"}

        writer = df.write.format(self._data_format).mode(mode)
        if mergeSchema is not None:
            writer = writer.option("mergeSchema", "true" if mergeSchema else "false")

        if overwriteSchema is not None:
            writer = writer.option(
                "overwriteSchema", "true" if overwriteSchema else "false"
            )

        if self._location:
            return writer.save(self._location)

        return writer.saveAsTable(self._name)

    def overwrite(
        self, df: DataFrame, mergeSchema: bool = None, overwriteSchema: bool = None
    ) -> None:
        return self.write_or_append(
            df, "overwrite", mergeSchema=mergeSchema, overwriteSchema=overwriteSchema
        )

    def append(self, df: DataFrame, mergeSchema: bool = None) -> None:
        return self.write_or_append(df, "append", mergeSchema=mergeSchema)

    def truncate(self) -> None:
        Spark.get().sql(f"TRUNCATE TABLE {self._name};")

    def drop(self) -> None:
        Spark.get().sql(f"DROP TABLE IF EXISTS {self._name};")

    def drop_and_delete(self) -> None:
        self.drop()
        if self._location:
            init_dbutils().fs.rm(self._location, True)

    def create_hive_table(self) -> None:
        sql = f"CREATE TABLE IF NOT EXISTS {self._name} "
        if self._location:
            sql += f" USING DELTA LOCATION '{self._location}'"
        Spark.get().sql(sql)

    def recreate_hive_table(self):
        self.drop()
        self.create_hive_table()

    def get_partitioning(self):
        """The result of DESCRIBE DETAIL tablename is like this:
        +------+--------------------+--------------------+----------------+-------+
        |format|                  id|                name|partitionColumns|  ...  |
        +------+--------------------+--------------------+----------------+-------+
        | delta|c96a1e94-314b-427...|spark_catalog.tes...|    [colB, colA]|  ...  |
        +------+--------------------+--------------------+----------------+-------+
        but this method return the partitioning in the form ['mycolA'],
        if there is no partitioning, an empty list is returned.
        """
        if self._partitioning is None:
            self._partitioning = (
                Spark.get()
                .sql(f"DESCRIBE DETAIL {self.get_tablename()}")
                .select("partitionColumns")
                .collect()[0][0]
            )
        return self._partitioning

    def get_tablename(self) -> str:
        return self._name

    def upsert(
        self,
        df: DataFrame,
        join_cols: List[str],
    ) -> Union[DataFrame, None]:
        if df is None:
            return None

        df = df.filter(" AND ".join(f"({col} is NOT NULL)" for col in join_cols))
        print(
            "Rows with NULL join keys found in input dataframe"
            " will be discarded before load."
        )

        df_target = self.read()

        # If the target is empty, always do faster full load
        if len(df_target.take(1)) == 0:
            return self.write_or_append(df, mode="overwrite")

        # Find records that need to be updated in the target (happens seldom)

        # Define the column to be used for checking for new rows
        # Checking the null-ness of one right row is sufficient to mark the row as new,
        # since null keys are disallowed.

        df, merge_required = CheckDfMerge(
            df=df,
            df_target=df_target,
            join_cols=join_cols,
            avoid_cols=[],
        )

        if not merge_required:
            return self.write_or_append(df, mode="append")

        temp_view_name = get_unique_tempview_name()
        df.createOrReplaceGlobalTempView(temp_view_name)

        target_table_name = self.get_tablename()
        non_join_cols = [col for col in df.columns if col not in join_cols]

        merge_sql_statement = GetMergeStatement(
            merge_statement_type="delta",
            target_table_name=target_table_name,
            source_table_name="global_temp." + temp_view_name,
            join_cols=join_cols,
            insert_cols=df.columns,
            update_cols=non_join_cols,
            special_update_set="",
        )

        df._jdf.sparkSession().sql(merge_sql_statement)

        print("Incremental Base - incremental load with merge")

        return df

    def delete_data(
        self, comparison_col: str, comparison_limit: Any, comparison_operator: str
    ) -> None:
        needs_quotes = (
            isinstance(comparison_limit, str)
            or isinstance(comparison_limit, date)
            or isinstance(comparison_limit, datetime)
        )
        limit = f"'{comparison_limit}'" if needs_quotes else comparison_limit
        sql_str = (
            f"DELETE FROM {self._name}"
            f" WHERE {comparison_col} {comparison_operator} {limit};"
        )
        Spark.get().sql(sql_str)

    def read_stream(self) -> DataFrame:
        reader = (
            Spark.get()
            .readStream.format(self._data_format)
            .options(**self._options_dict)
        )
        if self._location:
            df = reader.load(self._location)
        else:
            df = reader.table(self._table_name)

        return df

    def set_options_dict(self, options: Dict[str, str]):
        self._options_dict = options



import argparse

from spetlr.exceptions.cli_exceptions import SpetlrCliException

from . import generate_keys_file


class ConfiguratorCli:
    """This base class encapsulates the cli functions of the configurator.
    It is not intended to be instantiated directly."""

    def cli(self):
        """
        This function exposes functionality to be used in code analysis and deployment.
        Initialize the configurator in your project like

        ```
        def init_my_configurator():
            c= Configurator()
            c.add_resource_path(my_yaml_module)
            c.register('ENV', config.my_env_name)
            return c

        def my_configurator_cli():
            init_my_configurator().cli()
        ```
        Then expose the `my_configurator_cli` as a command line tool.
        The cli has the following usage

        <my_cli> [action [action_options]]

        Available actions:

        - generate-keys-file [-o output_file]
            Generates a python file that contains every available key string as a python
            object of the same name. Including this file in your python project allows
            you to get auto-completion of key names.
            The output is dumped to stdout unless -o output_file is specified.
            The option -o has the side-effect of return an exit code 1 if the file was
            updated. This allows you to check for a correctly updated keys file in your
            CICD pipleine.
        """
        parser = argparse.ArgumentParser()
        subp = parser.add_subparsers(help="actions")

        generate_keys_file.setup_parser(subp.add_parser("generate-keys-file"))
        # add further parsers here

        options = parser.parse_args()

        try:
            func = options.func
        except AttributeError:
            parser.print_help()
            raise SpetlrCliException("No actions given")

        func(options)
